import pytest

from eunomia import eunomia_load
from eunomia.backend import BackendDict
from eunomia.config import Group, Option
from eunomia.config.nodes import SubNode
from eunomia.registry import RegistryGroup
from eunomia.util._util_dict import recursive_setitem


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def foo(bar, baz=1):  # pragma: no cover
    return bar, baz


def test_simple_option():
    assert RegistryGroup().register_target(foo) is foo    # @register
    assert RegistryGroup().register_target()(foo) is foo  # @register()

    root = RegistryGroup()

    _foo = foo
    _foo = root.register_target()(_foo)
    _foo = root.register_target(name='foo_alt')(_foo)
    _foo = root.register_target(name='foo_alt2', path='/group/subgroup')(_foo)
    _foo = root.register_target(name='foo_alt3', path='/group/subgroup', params=dict(baz=2))(_foo)
    assert _foo is foo

    target = Group({
        'tests': Group({
            'test_registry': Group({
                'foo': Option({'_target_': 'tests.test_registry.foo', 'baz': 1}),
                'foo_alt': Option({'_target_': 'tests.test_registry.foo', 'baz': 1}),
            })
        }),
        'group': Group({
            'subgroup': Group({
                'foo_alt2': Option({'_target_': 'tests.test_registry.foo', 'baz': 1}),
                'foo_alt3': Option({'_target_': 'tests.test_registry.foo', 'baz': 2}),
            })
        })
    })

    # NOTE: the working directory must not be /eunomia/tests, it must be /eunomia
    assert BackendDict().dump(root) == BackendDict().dump(target)

    # test defaults
    # - if is_default is not specified, then the default can be overwritten
    #   otherwise it cannot be

    assert root.get_registered_defaults() == [{'/tests/test_registry': 'foo'}]
    assert root.get_registered_defaults(explicit_only=True) == []

    _foo = root.register_target(name='foo_alt4', path='/group/subgroup', params=dict(baz=3), is_default=True)(_foo)
    # this should fail because it was also marked as the default
    with pytest.raises(AssertionError, match='registered callable .* for option: .* was previously explicitly registered as a default.'):
        root.register_target(name='foo_alt4_CHECK', path='/group/subgroup', params=dict(baz=3), is_default=True)(_foo)
    # this should not fail as the above failed
    root.register_target(name='foo_alt4_CHECK', path='/group/subgroup', params=dict(baz=3), is_default=False)(_foo)

    assert root.get_registered_defaults() == [{'/group/subgroup': 'foo_alt4'}]
    assert root.get_registered_defaults(explicit_only=True) == [{'/group/subgroup': 'foo_alt4'}]

    # test that the registrable group is the root.
    g = Group({'temp': root})
    with pytest.raises(AssertionError, match='Can only register on the root node.'): root.register_target()(_foo)
    with pytest.raises(AssertionError, match='Can only register on the root node.'): root.get_registered_defaults()

    # reset
    root._parent = None
    root._key = None

    # try again
    # TODO: maybe update to better error message
    with pytest.raises(KeyError, match='parent already has child with key: foo'): root.register_target()(_foo)
    assert root.get_registered_defaults() == [{'/group/subgroup': 'foo_alt4'}]

    # try merge group
    root.add_option('default', Option(defaults=root.get_registered_defaults()))
    assert eunomia_load(root) == {'group': {'subgroup': {'_target_': 'tests.test_registry.foo', 'baz': 3}}}


def fizz(foo, bar, buzz=1, baz=1):  # pragma: no cover
    return foo, bar, buzz, baz


def test_override_modes():
    def _register(**kwargs):
        root = RegistryGroup()
        root.register_target_fn(fizz, **kwargs)
        root.add_option('default', Option(defaults=root.get_registered_defaults()))
        return eunomia_load(root)

    with pytest.raises(KeyError, match='Invalid override mode: \'INVALID\''):
        _register(mode='INVALID')

    # should be any
    assert _register() == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}
    assert _register(keep_defaults=True)  == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}
    assert _register(keep_defaults=False) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz'}}}
    assert _register(keep_defaults=True, params=dict(baz=2)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 2}}}
    assert _register(keep_defaults=False, params=dict(baz=2)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'baz': 2}}}

    # mode: any
    assert _register(keep_defaults=True, mode='any')  == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}
    assert _register(keep_defaults=False, mode='any') == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz'}}}
    assert _register(keep_defaults=True, mode='any', params=dict(buzz=5))  == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 5, 'baz': 1}}}
    assert _register(keep_defaults=False, mode='any', params=dict(buzz=5)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 5}}}
    assert _register(keep_defaults=False, mode='any', params=dict(foo=7)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'foo': 7}}}
    assert _register(keep_defaults=True, mode='any', params=dict(foo=7, bar=7)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'foo': 7, 'bar': 7, 'baz': 1, 'buzz': 1}}}
    assert _register(keep_defaults=True, mode='any', params=dict(foo=7, bar=7, buzz=5)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'foo': 7, 'bar': 7, 'baz': 1, 'buzz': 5}}}
    with pytest.raises(AssertionError, match="cannot override params: \\['bazz']"):
        _register(keep_defaults=True, mode='any', params=dict(foo=7, bar=7, buzz=5, bazz=-1))

    # mode: kwargs
    assert _register(keep_defaults=True, mode='kwargs')  == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}
    assert _register(keep_defaults=False, mode='kwargs') == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz'}}}
    assert _register(keep_defaults=True, mode='kwargs', params=dict(buzz=5))  == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 5, 'baz': 1}}}
    assert _register(keep_defaults=False, mode='kwargs', params=dict(buzz=5)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 5}}}
    with pytest.raises(AssertionError, match="cannot override params: \\['foo']"):
        _register(keep_defaults=False, mode='kwargs', params=dict(foo=7))
    with pytest.raises(AssertionError, match="cannot override params: \\['foo', 'bar']"):
        _register(keep_defaults=True, mode='kwargs', params=dict(foo=7, bar=7))
    with pytest.raises(AssertionError, match="cannot override params: \\['foo', 'bar']"):
        _register(keep_defaults=True, mode='kwargs', params=dict(foo=7, bar=7, buzz=5))

    # made: all
    with pytest.raises(AssertionError, match="all non-default parameters require an override: \\['foo', 'bar']"):
        _register(keep_defaults=True, mode='all')
    with pytest.raises(AssertionError, match="all non-default parameters require an override: \\['foo', 'bar']"):
        _register(keep_defaults=False, mode='all')
    with pytest.raises(AssertionError, match="all non-default parameters require an override: \\['foo']"):
        _register(keep_defaults=True, mode='all', params=dict(bar=7))
    with pytest.raises(AssertionError, match="all non-default parameters require an override: \\['bar']"):
        _register(keep_defaults=False, mode='all', params=dict(foo=7))
    assert _register(keep_defaults=False, mode='all', params=dict(foo=7, bar=7)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'bar': 7, 'foo': 7}}}
    assert _register(keep_defaults=True, mode='all', params=dict(foo=7, bar=7)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'bar': 7, 'foo': 7, 'buzz': 1, 'baz': 1}}}

    # made: unchecked
    assert _register(keep_defaults=True, mode='unchecked') == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'baz': 1, 'buzz': 1}}}
    assert _register(keep_defaults=False, mode='unchecked') == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz'}}}
    assert _register(keep_defaults=True, mode='unchecked', params=dict(INVALID=1)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'baz': 1, 'buzz': 1, 'INVALID': 1}}}
    assert _register(keep_defaults=False, mode='unchecked', params=dict(INVALID=1)) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'INVALID': 1}}}

    # test nesting inside option and extra data
    assert _register() == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}
    assert _register(nest_path='targ1') == {'tests': {'test_registry': {'targ1': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}}
    assert _register(nest_path='targ1.targ2') == {'tests': {'test_registry': {'targ1': {'targ2': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}}}
    assert _register(nest_path='targ1', data={'asdf': 1}) == {'tests': {'test_registry': {'asdf': 1, 'targ1': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}}
    assert _register(nest_path='targ1', data={'targ1': {}}) == {'tests': {'test_registry': {'targ1': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}}}
    with pytest.raises(ValueError, match='nested object in data must be empty, otherwise target conflicts can occur'):
        _register(nest_path='targ1', data={'targ1': {'foo': 1}})
    with pytest.raises(ValueError, match='nested object in data must be a dictionary that the target can be merged into'):
        _register(nest_path='targ1', data={'targ1': []})
    assert _register(nest_path='targ1', pkg='<root>') == {'targ1': {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}}
    assert _register(pkg='<root>') == {'_target_': 'tests.test_registry.fizz', 'buzz': 1, 'baz': 1}



def test_registry_with_options_as_defaults():

    registry = RegistryGroup()
    option_foo = registry.register_target_fn(foo, params=dict(bar=20))
    option_fizz = registry.register_target_fn(fizz, params=dict(foo=10, bar=20))

    registry.add_option('default', Option(defaults=[option_foo]))
    assert eunomia_load(registry) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.foo', 'bar': 20, 'baz': 1}}}

    with pytest.raises(KeyError, match='parent already has child with key'):
        registry.add_option('default', Option(defaults=[option_foo, option_fizz]))

    registry.del_option('default')
    registry.add_option('default', Option(defaults=[option_fizz]))
    assert eunomia_load(registry) == {'tests': {'test_registry': {'_target_': 'tests.test_registry.fizz', 'foo': 10, 'bar': 20, 'buzz': 1, 'baz': 1}}}


def fn(name='N/A'):
    return name


def test_registry_with_options_as_defaults_advanced():
    registry = RegistryGroup()

    # data type -- what kind of data is being used, this affects what type of data wrapper is needed for the framework
    data_type = registry.get_group_recursive('auto/data_type', make_missing=True)
    type_ground_truth = data_type.new_option('ground_truth', pkg='auto', data=dict(data_type='ground_truth'))
    type_episodes     = data_type.new_option('episodes',     pkg='auto', data=dict(data_type='episodes'))

    # data wrap mode -- how many inputs needs to be sampled to form a single observation from the dataset
    data_wrap_mode = registry.get_group_recursive('auto/data_wrap_mode', make_missing=True)
    wrap_triples = data_wrap_mode.new_option('triples', pkg='auto', data=dict(data_wrap_mode='triples'), defaults=[SubNode('/auto/data_wrapper/${auto.data_type}_${auto.data_wrap_mode}')])
    wrap_pairs   = data_wrap_mode.new_option('pairs',   pkg='auto', data=dict(data_wrap_mode='pairs'),   defaults=[SubNode('/auto/data_wrapper/${auto.data_type}_${auto.data_wrap_mode}')])
    wrap_single  = data_wrap_mode.new_option('single',  pkg='auto', data=dict(data_wrap_mode='single'),  defaults=[SubNode('/auto/data_wrapper/${auto.data_type}_${auto.data_wrap_mode}')])

    # dataset wrappers
    registry.register_target(fn, name='ground_truth_single',     params=dict(name='ground_truth_single'),     path='auto/data_wrapper')
    registry.register_target(fn, name='ground_truth_pairs',      params=dict(name='ground_truth_pairs'),      path='auto/data_wrapper')
    registry.register_target(fn, name='ground_truth_weak_pairs', params=dict(name='ground_truth_weak_pairs'), path='auto/data_wrapper')
    registry.register_target(fn, name='ground_truth_triples',    params=dict(name='ground_truth_triples'),    path='auto/data_wrapper')
    registry.register_target(fn, name='episodes_single',         params=dict(name='episodes_single'),         path='auto/data_wrapper')  # num_samples=1, sample_radius=32))
    registry.register_target(fn, name='episodes_pairs',          params=dict(name='episodes_pairs'),          path='auto/data_wrapper')  # num_samples=2, sample_radius=32))
    registry.register_target(fn, name='episodes_triples',        params=dict(name='episodes_triples'),        path='auto/data_wrapper')  # num_samples=3, sample_radius=32))

    # data
    registry.register_target(fn, name='dsprites',       params=dict(name='dsprites'),       path='disent/data', defaults=[type_ground_truth])
    registry.register_target(fn, name='monte_rollouts', params=dict(name='monte_rollouts'), path='disent/data', defaults=[type_episodes])

    # frameworks
    registry.register_target(fn, name='betavae', params=dict(name='betavae'), path='disent/framework', nest_path='cfg', defaults=[wrap_single])
    registry.register_target(fn, name='adavae',  params=dict(name='adavae'),  path='disent/framework', nest_path='cfg', defaults=[wrap_pairs])
    registry.register_target(fn, name='tvae',    params=dict(name='tvae'),    path='disent/framework', nest_path='cfg', defaults=[wrap_triples])

    def check(in_data, in_framework, out_data_type, out_data_wrap_mode):
        if registry.has_suboption('default'):
            registry.del_option('default')

        registry.new_option('default', defaults=[
            {'/disent/data': in_data},
            {'/disent/framework': in_framework},
        ])

        assert eunomia_load(registry) == {'disent': {'data': {'_target_': 'tests.test_registry.fn', 'name': in_data}, 'framework': {'cfg': {'_target_': 'tests.test_registry.fn', 'name': in_framework}}}, 'auto': {'data_type': out_data_type, 'data_wrap_mode': out_data_wrap_mode, 'data_wrapper': {'_target_': 'tests.test_registry.fn', 'name': f'{out_data_type}_{out_data_wrap_mode}'}}}

    check('dsprites', 'betavae', 'ground_truth', 'single')
    check('dsprites', 'adavae', 'ground_truth', 'pairs')
    check('dsprites', 'tvae', 'ground_truth', 'triples')

    check('monte_rollouts', 'betavae', 'episodes', 'single')
    check('monte_rollouts', 'adavae', 'episodes', 'pairs')
    check('monte_rollouts', 'tvae', 'episodes', 'triples')

    # CHECK MULTIPLE OPTIONS IN A SINGLE DEFAULT

    option_metric_dci = registry.register_target_fn(fn, name='metric_dci', params=dict(name='metric_dci'), pkg='<option>', path='disent/metrics')
    option_metric_mig = registry.register_target_fn(fn, name='metric_mig', params=dict(name='metric_mig'), pkg='<option>', path='disent/metrics')
    option_metric_fvs = registry.register_target_fn(fn, name='metric_fvs', params=dict(name='metric_fvs'), pkg='<option>', path='disent/metrics')

    def check(defaults, out_metrics, pkg_keys=None):
        if registry.has_suboption('default'):
            registry.del_option('default')

        registry.new_option('default', defaults=defaults)

        output = {}
        merged_metrics = {
            metric: {'_target_': 'tests.test_registry.fn', 'name': metric}
            for metric in out_metrics
        }
        recursive_setitem(output, ['disent', 'metrics'], merged_metrics, make_missing=True)

        assert eunomia_load(registry) == output

    # check that we can load all of them
    check(defaults=['/disent/metrics/metric_dci'], out_metrics=['metric_dci'])
    check(defaults=[option_metric_dci], out_metrics=['metric_dci'])

    check(defaults=[{'/disent/metrics': ['/disent/metrics/metric_dci', '/disent/metrics/metric_mig']}], out_metrics=['metric_dci', 'metric_mig'])
    check(defaults=[{'/disent/metrics': ['/disent/metrics/metric_dci', option_metric_mig]}], out_metrics=['metric_dci', 'metric_mig'])
    check(defaults=[{'/disent/metrics': '*'}], out_metrics=['metric_dci', 'metric_mig', 'metric_fvs'])

    check(defaults=[{'/disent/metrics': [option_metric_dci, option_metric_mig]}], out_metrics=['metric_dci', 'metric_mig'])
    check(defaults=[{'/disent/metrics': [option_metric_dci, option_metric_mig, option_metric_fvs]}], out_metrics=['metric_dci', 'metric_mig', 'metric_fvs'])

    # check that the group can be added itself
    check([registry.get_group_recursive('/disent/metrics')], ['metric_dci', 'metric_mig', 'metric_fvs'])



# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from eunomia.backend import Backend
from eunomia.config import Group, Option


# ========================================================================= #
# Object Backend                                                            #
# ========================================================================= #


class BackendObj(Backend):
    """
    The most basic backend is the ConfigObject data structures
    used by the core config loader.

    The backend passes through the values as they
    are without any modifications.
    """

    GROUP_TYPE = Group
    OPTION_TYPE = Option

    def _load_group(cls, value) -> Group:
        return value  # pragma: no cover

    def _load_option(cls, value) -> Option:
        return value  # pragma: no cover

    def _dump_group(self, group: Group):
        return group  # pragma: no cover

    def _dump_option(self, option: Option):
        return option  # pragma: no cover


# ========================================================================= #
# END                                                                       #
# ========================================================================= #





# ========================================================================= #
# Object Traversal                                                          #
# ========================================================================= #


class Transformer(object):
    """
    Based off of the various transformers that
    exist for traversing ASTs.
    """

    def transform(self, value):
        attr = f'_transform_{type(value).__name__}'
        func = getattr(self, attr, self.__transform_default__)
        return func(value)

    def __transform_default__(self, value):
        raise NotImplementedError


class RecursiveTransformer(Transformer):
    """
    This one effectively makes copies of the
    containers being traversed.
    """

    def _transform_set(self, value):
        return set(self.transform(v) for v in value)

    def _transform_list(self, value):
        return list(self.transform(v) for v in value)

    def _transform_tuple(self, value):
        return tuple(self.transform(v) for v in value)

    def _transform_dict(self, value):
        return {self._transform_dict_key(k): self._transform_dict_value(v) for k, v in value.items()}

    def _transform_dict_key(self, key):
        return self.transform(key)

    def _transform_dict_value(self, value):
        return self.transform(value)


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import pytest

from eunomia.backend import BackendDict
from eunomia.config import Group, Option


# ========================================================================= #
# Test Eunomia                                                              #
# ========================================================================= #


def test_config_objects():

    bk = BackendDict()

    left = Group(dict(
        suboption1=Option(dict(foo=1)),
        subgroup1=Group(dict(
            suboption1=Option(dict(foo=1)),
        )),
    ))

    right = Group(dict(
        suboption2=Option(dict(foo=2)),
        subgroup1=Group(dict(
            suboption2=Option(dict(foo=2)),
        )),
        subgroup2=Group(dict(
            suboption1=Option(dict(foo=1)),
            suboption2=Option(dict(foo=2)),
        )),
    ))

    # absorb
    left.absorb_children(right)
    with pytest.raises(TypeError, match='group can only absorb other groups'):
        left.absorb_children(Option())
    with pytest.raises(AssertionError, match='tried to absorb group from the same tree'):
        left.absorb_children(left.get_subgroup('subgroup1'))

    target = Group(dict(
        suboption1=Option(dict(foo=1)),
        suboption2=Option(dict(foo=2)),
        subgroup1=Group(dict(
            suboption1=Option(dict(foo=1)),
            suboption2=Option(dict(foo=2)),
        )),
        subgroup2=Group(dict(
            suboption1=Option(dict(foo=1)),
            suboption2=Option(dict(foo=2)),
        )),
    ))

    # checks
    assert bk.dump(left) == bk.dump(target)
    assert bk.dump(right) == bk.dump(Group())

    right2 = Group(dict(
        subgroup1=Group(dict(
            suboption1=Option(dict(foo=1)),
        )),
    ))

    target2 = Group(dict(
        suboption1=Option(dict(foo=1)),
        suboption2=Option(dict(foo=2)),
        subgroup1=Group(dict(
            suboption1=Option(dict(foo=1)),
            suboption2=Option(dict(foo=2)),
        )),
        subgroup2=Group(dict(
            suboption1=Option(dict(foo=1)),
            suboption2=Option(dict(foo=2)),
            subgroup1=Group(dict(
                suboption1=Option(dict(foo=1)),
            )),
        )),
    ))

    left.get_subgroup('subgroup2').absorb_children(right2)

    assert bk.dump(left) == bk.dump(target2)
    assert bk.dump(right) == bk.dump(Group())

    right3 = Group(dict(
        subgroup2=Group(dict(
            suboption1=Option(dict(foo=111)),
            subgroup1=Group(dict(
                suboption1=Option(dict(foo=111)),
            )),
        )),
    ))

    target3 = Group(dict(
        suboption1=Option(dict(foo=1)),
        suboption2=Option(dict(foo=2)),
        subgroup1=Group(dict(
            suboption1=Option(dict(foo=1)),
            suboption2=Option(dict(foo=2)),
        )),
        subgroup2=Group(dict(
            suboption1=Option(dict(foo=111)),
            suboption2=Option(dict(foo=2)),
            subgroup1=Group(dict(
                suboption1=Option(dict(foo=111)),
            )),
        )),
    ))

    # check conflict, both roots
    with pytest.raises(KeyError, match="left group '/' cannot absorb right group '/' because conflicting left option was found"):
        left.absorb_children(right3)
    # check conflict, right is child
    _right3_root = Group(dict(parentgroup=right3))
    with pytest.raises(KeyError, match="left group '/' cannot absorb right group '/parentgroup' because conflicting left option was found"):
        left.absorb_children(right3)
    assert bk.dump(left) == bk.dump(target2)
    # check conflict, left is child and right is child
    _left_root = Group(dict(parentgroup=left))
    with pytest.raises(KeyError, match="left group '/parentgroup' cannot absorb right group '/parentgroup' because conflicting left option was found"):
        left.absorb_children(right3)
    assert bk.dump(left) == bk.dump(target2)
    # check conflict, left is child
    _right3_root.del_subgroup('parentgroup')
    with pytest.raises(KeyError, match="left group '/parentgroup' cannot absorb right group '/' because conflicting left option was found"):
        left.absorb_children(right3)
    assert bk.dump(left) == bk.dump(target2)
    # check conflict, both roots, AGAIN
    _left_root.del_subgroup('parentgroup')
    with pytest.raises(KeyError, match="left group '/' cannot absorb right group '/' because conflicting left option was found"):
        left.absorb_children(right3)
    assert bk.dump(left) == bk.dump(target2)

    # merge with overwrite enabled
    left.absorb_children(right3, allow_replace_options=True)
    assert bk.dump(left) == bk.dump(target3)



import pytest

from eunomia.backend import BackendDict
from eunomia.config import Group, Option
from eunomia.config import keys as K


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def test_simple_option():
    bk = BackendDict()

    assert bk.dump_option(Option(data={'foo': 'bar'})) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': '<group>', '__type__': 'option'}
    assert bk.dump_option(Option(data={'foo': 'bar'}, pkg=K.PKG_ROOT)) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': '<root>', '__type__': 'option'}
    assert bk.dump_option(Option(data={'foo': 'bar'}, pkg=K.PKG_GROUP)) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': '<group>', '__type__': 'option'}
    assert bk.dump_option(Option(data={'foo': 'bar'}, pkg=K.PKG_OPTION)) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': '<option>', '__type__': 'option'}

    assert bk.dump_option(Option(data={'foo': 'bar'}, defaults=[])) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': '<group>', '__type__': 'option'}

    # test relative path
    assert bk.dump_option(Option(data={'foo': 'bar'}, defaults=['group1/option1'])) == {'__data__': {'foo': 'bar'}, '__defaults__': ['group1/option1'], '__package__': '<group>', '__type__': 'option'}
    assert bk.dump_option(Option(data={'foo': 'bar'}, defaults=['group1/group2/option2'])) == {'__data__': {'foo': 'bar'}, '__defaults__': ['group1/group2/option2'], '__package__': '<group>', '__type__': 'option'}
    # test absolute path
    assert bk.dump_option(Option(data={'foo': 'bar'}, defaults=['/group1/group2/option2'])) == {'__data__': {'foo': 'bar'}, '__defaults__': ['/group1/group2/option2'], '__package__': '<group>', '__type__': 'option'}

    with pytest.raises(ValueError):  # TODO: fix error messages, match='is_identifier'):
        bk.dump_option(Option(data={'foo': 'bar'}, defaults=[{'group1': '1invalid'}]))
    with pytest.raises(ValueError):  # TODO: fix error messages, match='Wrong key'):
        bk.dump_option(Option(data={'foo': 'bar'}, defaults=[{'1invalid': 'option2'}]))
    with pytest.raises(ValueError):  # TODO: fix error messages, match='Wrong key'):
        bk.dump_option(Option(data={'foo': 'bar'}, defaults=[{'group1/2invalid': 'option2'}]))

    assert bk.dump_option(Option(data={'foo': 'bar'}, pkg='key1')) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': 'key1', '__type__': 'option'}
    assert bk.dump_option(Option(data={'foo': 'bar'}, pkg='key1.key2')) == {'__data__': {'foo': 'bar'}, '__defaults__': [], '__package__': 'key1.key2', '__type__': 'option'}


def test_simple_config():
    g = Group({
        'subgroup': Group({
            'suboption2': Option({'bar': 2}),
            'suboption1': Option({'bar': 1}),
        }),
        'subgroup2': Group({
            'sub2option1': Option({'baz': 1}),
            'sub2option2': Option({'baz': 2}),
        }),
        'default': Option({
            'foo': 1
        }),
    })

    bck = BackendDict()

    # forward and backwards conversion!
    d1 = BackendDict().dump(g)
    g1 = BackendDict().load_group(d1)
    assert repr(g1) == repr(g)


def _make_config_group(suboption='suboption1', suboption2=None, package1='<group>', package2='<group>', check_cycle=False) -> Group:
    g = Group({
        'subgroup': Group({
            'suboption1': Option({'bar': 1}, pkg=package1),
            'suboption2': Option({'bar': 2}, pkg=package1),
        }),
        'subgroup2': Group({
            'subgroup3': Group({
                'suboption1': Option({'baz': 1}, pkg=package2),
                'suboption2': Option({'baz': 2}, pkg=package2),
            }),
        }),
        'default': Option(
            data={'foo': 1},
            defaults=[
                *([{'subgroup': suboption}] if suboption else []),  # relative
                *([{'/subgroup2/subgroup3': suboption2}] if suboption2 else []),  # absolute
            ],
        ),
    })

    if check_cycle:
        # forward and backwards conversion!
        d1 = BackendDict().dump(g)
        g1 = BackendDict().load_group(d1)
        assert repr(g1) == repr(g)

    return g


def test_config_objects():
    _make_config_group(suboption='suboption1', check_cycle=True)
    _make_config_group(suboption='suboption2', check_cycle=True)
    _make_config_group(suboption='invalid___', check_cycle=True)  # invalid but not an error


def test_option_init():
    # raw data in option
    data = {'foo': 'bar'}
    defaults = ['/group1/group2/option2']
    pkg = K.PKG_ROOT
    # make equivalent options
    option = Option(data=data, defaults=defaults, pkg=pkg)
    # check circular conversion
    bk = BackendDict()
    assert bk.dump(option) == bk.dump(bk.load_option(bk.dump(option)))


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from eunomia.config.nodes import SubNode


def test_node_sub_fstrings():
    assert SubNode('f""').get_config_value({}, {}, {}) == ''
    assert SubNode('f"{1}"').get_config_value({}, {}, {}) == '1'
    assert SubNode('f"{1}{[1,2]}"').get_config_value({}, {}, {}) == '1[1, 2]'
    assert SubNode('f"{1}{[1,2]*2}"').get_config_value({}, {}, {}) == '1[1, 2, 1, 2]'
    assert SubNode('f"| {\'=\'*5} |"').get_config_value({}, {}, {}) == '| ===== |'
    assert SubNode('f"| {1+2+0.56:1.1f} |"').get_config_value({}, {}, {}) == '| 3.6 |'
    assert SubNode('f"{round(0.5)}"').get_config_value({}, {}, {}) == '0'

def test_node_sub():
    assert SubNode('a{1}b{2}c').get_config_value({},{},{}) == 'a{1}b{2}c'
    assert SubNode('a${=1}b${=2}c').get_config_value({},{},{}) == 'a1b2c'



import itertools

import pytest

from eunomia.backend import BackendDict
from eunomia.config import Group, Option
from eunomia.util._targets import call, instantiate, import_module_attr
import ruamel
import math


def test_import_objects():
    assert import_module_attr('dict') is dict
    assert import_module_attr('list') is list

    assert import_module_attr('math.log') is math.log

    assert import_module_attr('ruamel.yaml') is ruamel.yaml  # odd one
    assert import_module_attr('ruamel.yaml.round_trip_dump') is ruamel.yaml.round_trip_dump  # odd one

    assert import_module_attr('eunomia.config.Group') is Group
    assert import_module_attr('eunomia.config.Option') is Option
    assert import_module_attr('eunomia.config.Group.has_suboption') is Group.has_suboption
    assert import_module_attr('eunomia.config.Option.get_child') is Option.get_child

    with pytest.raises(ValueError, match='must contain at least one module component'):
        import_module_attr('')
    with pytest.raises(AttributeError, match="could not import target 'math'. module 'builtins' has no attribute 'math'"):
        import_module_attr('math')
    with pytest.raises(AttributeError, match="could not import target 'math.asdf'. module 'math' has no attribute 'asdf'"):
        import_module_attr('math.asdf')
    with pytest.raises(AttributeError, match="could not import target 'math.log.asdf'. 'builtin_function_or_method' object has no attribute 'asdf'"):
        import_module_attr('math.log.asdf')
    with pytest.raises(AttributeError, match="could not import target 'eunomia.config.Option.asdf'. type object 'Option' has no attribute 'asdf'"):
        import_module_attr('eunomia.config.Option.asdf')
    with pytest.raises(ImportError, match="could not import target 'eunomiaINVALID.config.Option.get_child'. Are you sure the import module is correct?"):
        import_module_attr('eunomiaINVALID.config.Option.get_child')

def test_instantiate_targets():
    # check that the functions are the same, just nice to indicate whats going on
    assert call is instantiate

    # check basic instantiate
    assert isinstance(instantiate({'_target_': 'eunomia.config.Group'}, recursive=True, ensure_root_target=True), Group)
    assert isinstance(instantiate({'_target_': 'eunomia.config.Option'}, recursive=True, ensure_root_target=False), Option)
    assert isinstance(instantiate({'_target_': 'eunomia.config.Group'}, recursive=False, ensure_root_target=True), Group)
    assert isinstance(instantiate({'_target_': 'eunomia.config.Option'}, recursive=False, ensure_root_target=False), Option)

    bk = BackendDict()
    # check recursive instantiate
    option = Option(data={'foo': 1}, pkg='pkg.subpkg', defaults=['group/subgroup'])
    option_target = {'_target_': 'eunomia.config.Option', '_args_': [dict(foo=1), 'pkg.subpkg', ['group/subgroup']]}
    # check no recursive elements
    assert bk.dump(instantiate(option_target, recursive=True, ensure_root_target=True)) == bk.dump(option)
    assert bk.dump(instantiate(option_target, recursive=True, ensure_root_target=False)) == bk.dump(option)
    assert bk.dump(instantiate(option_target, recursive=False, ensure_root_target=True)) == bk.dump(option)
    assert bk.dump(instantiate(option_target, recursive=False, ensure_root_target=False)) == bk.dump(option)
    # check wrong base element
    # -- no recursive
    assert instantiate([option_target], recursive=False, ensure_root_target=False) == [option_target]
    assert instantiate({'asdf': option_target}, recursive=False, ensure_root_target=False) == {'asdf': option_target}
    # -- no recursive -- strict
    with pytest.raises(TypeError, match='Could not instantiate root object: not a dictionary'):
        instantiate([option_target], recursive=False, ensure_root_target=True)
    with pytest.raises(KeyError, match="Could not instantiate root object: target marker key '_target_' not found in root"):
        instantiate({'asdf': option_target}, recursive=False, ensure_root_target=True)

    # check advanced recursive instantiate
    group = Group({'default': option})
    group_target = {'_target_': 'eunomia.config.Group', '_args_': [{'default': option_target}]}
    group_target_WRONG = {'_target_': 'eunomia.config.Option', '_args_': [{'default': option_target}]}
    # checks wrong call!
    with pytest.raises(RuntimeError, match="Could not call target 'eunomia.config.Option' object <class"):
        instantiate(group_target_WRONG, recursive=True, ensure_root_target=True)
    # checks recursive call
    assert bk.dump(instantiate(group_target, recursive=True, ensure_root_target=True)) == bk.dump(group)

    # check advanced recursive instantiate
    option_alt = Option(data={'foo': 1}, pkg='pkg.subpkg', defaults=[{'group': 'subgroup'}])
    # args, kwargs key mix
    option_target_alt1 = {'_target_': 'eunomia.config.Option', '_args_': [dict(foo=1), 'pkg.subpkg', [{'group': 'subgroup'}]]}
    option_target_alt2 = {'_target_': 'eunomia.config.Option', '_args_': [dict(foo=1), 'pkg.subpkg'], '_kwargs_': dict(defaults=[{'group': 'subgroup'}])}
    option_target_alt3 = {'_target_': 'eunomia.config.Option', '_args_': [dict(foo=1)], '_kwargs_': dict(pkg='pkg.subpkg', defaults=[{'group': 'subgroup'}])}
    option_target_alt4 = {'_target_': 'eunomia.config.Option', '_kwargs_': dict(data=dict(foo=1), pkg='pkg.subpkg', defaults=[{'group': 'subgroup'}])}
    assert bk.dump(instantiate(option_target_alt1)) == bk.dump(option_alt)
    assert bk.dump(instantiate(option_target_alt2)) == bk.dump(option_alt)
    assert bk.dump(instantiate(option_target_alt3)) == bk.dump(option_alt)
    assert bk.dump(instantiate(option_target_alt4)) == bk.dump(option_alt)
    # args, root kwargs mix
    option_target_alt1 = {'_target_': 'eunomia.config.Option', '_args_': [dict(foo=1), 'pkg.subpkg'], 'defaults': [{'group': 'subgroup'}]}
    option_target_alt2 = {'_target_': 'eunomia.config.Option', '_args_': [dict(foo=1)], 'pkg': 'pkg.subpkg', 'defaults': [{'group': 'subgroup'}]}
    option_target_alt3 = {'_target_': 'eunomia.config.Option', 'data': dict(foo=1), 'pkg': 'pkg.subpkg', 'defaults': [{'group': 'subgroup'}]}
    option_target_alt4 = {'_target_': 'eunomia.config.Option', 'data': dict(foo=1), 'pkg': 'pkg.subpkg', '_kwargs_': dict(defaults=[{'group': 'subgroup'}])}
    assert bk.dump(instantiate(option_target_alt1)) == bk.dump(option_alt)
    assert bk.dump(instantiate(option_target_alt2)) == bk.dump(option_alt)
    assert bk.dump(instantiate(option_target_alt3)) == bk.dump(option_alt)
    with pytest.raises(KeyError, match="target dictionary contains the _kwargs_ key as well as unused keys"):
        assert bk.dump(instantiate(option_target_alt4)) == bk.dump(option_alt)



import pytest

from eunomia.config._default import Default
from eunomia.config.nodes import ConfigNode
from tests.test_backend_obj import _make_config_group


# ========================================================================= #
# Test YAML & Custom Tags                                                   #
# ========================================================================= #


def _resolver(string):
    if isinstance(string, ConfigNode):
        return string.get_config_value({}, {}, {})
    return string


def _resolve_default(group, default):
    # we are testing this! \/ \/ \/
    g, c, pkg, is_self = default.to_resolved_components(group, _resolver)
    # we are testing this! /\ /\ /\
    return g.abs_path, [c.abs_path for c in c], pkg


def test_defaults():

    root = _make_config_group(suboption=None, suboption2=None, package1='<option>', package2='asdf.fdsa')

    d = root.get_option_recursive('default')
    s1 = root.get_group_recursive('subgroup')
    s1o1 = root.get_option_recursive('subgroup/suboption1')
    s1o2 = root.get_option_recursive('subgroup/suboption2')
    s2 = root.get_group_recursive('subgroup2')
    s2s3 = root.get_group_recursive('subgroup2/subgroup3')
    s2s3o1 = root.get_option_recursive('subgroup2/subgroup3/suboption1')
    s2s3o2 = root.get_option_recursive('subgroup2/subgroup3/suboption2')

    # multiple different versions
    assert _resolve_default(root, Default(d))                   == ('/', ['/default'], ())
    assert _resolve_default(root, Default({root: d}))           == ('/', ['/default'], ())
    assert _resolve_default(root, Default({root: [d]}))         == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': 'default'}))    == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': '/default'}))   == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': ['/default']})) == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': ['default']}))  == ('/', ['/default'], ())

    # these should throw errors, option points to option
    with pytest.raises(KeyError, match='key .* is not a group'): _resolve_default(root, Default({'/default': ['default']}))
    with pytest.raises(KeyError, match='key .* is not a group'): _resolve_default(root, Default({d: d}))
    with pytest.raises(KeyError, match='key .* is not a group'): _resolve_default(root, Default({d: [d]}))

    # allow group to represent all suboptions
    assert _resolve_default(root, Default('')) == ('/', ['/default'], ())  # technically this is valid, its just confusing... should it be disabled?
    assert _resolve_default(root, Default('default')) == ('/', ['/default'], ())  # we want relative support in case we use group.absorb for example
    assert _resolve_default(root, Default('/')) == ('/', ['/default'], ())
    assert _resolve_default(root, Default('/default')) == ('/', ['/default'], ())
    assert _resolve_default(root, Default(root)) == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': '/'})) == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': '*'})) == ('/', ['/default'], ())
    assert _resolve_default(root, Default({root: '*'})) == ('/', ['/default'], ())
    assert _resolve_default(root, Default({root: root})) == ('/', ['/default'], ())
    assert _resolve_default(root, Default({'/': root})) == ('/', ['/default'], ())
    # these should throw errors, group points to group in list
    with pytest.raises(KeyError, match='value in list .* is not an option'): _resolve_default(root, Default({'/': ['subgroup']}))
    with pytest.raises(KeyError, match='value in list .* is not an option'): _resolve_default(root, Default({'/': ['default', 'subgroup']}))

    # check parents
    assert _resolve_default(root, Default(d)) == ('/', ['/default'], ())
    assert _resolve_default(s1, Default(d)) == ('/', ['/default'], ())
    assert _resolve_default(s2, Default(d)) == ('/', ['/default'], ())
    assert _resolve_default(s2s3, Default(d)) == ('/', ['/default'], ())

    # check others
    assert _resolve_default(root, Default(s1))        == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(s1,   Default(s1))        == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(s2,   Default(s1))        == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(s2s3, Default(s1))        == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(root, Default({s1: '*'})) == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(root, Default({s1: s1}))  == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))

    # strings
    assert _resolve_default(root, Default('subgroup'))                 == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(root, Default('/subgroup'))                == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(root, Default(s1))                         == ('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))
    assert _resolve_default(root, Default(s1o1))                       == ('/subgroup', ['/subgroup/suboption1'], ('subgroup',))
    assert _resolve_default(root, Default({'/subgroup': 'suboption1'})) == ('/subgroup', ['/subgroup/suboption1'], ('subgroup',))
    assert _resolve_default(root, Default({'subgroup': 'suboption1'}))  == ('/subgroup', ['/subgroup/suboption1'], ('subgroup',))
    with pytest.raises(KeyError, match="Group '/subgroup2/subgroup3' does not have child 'subgroup'"):
        _resolve_default(s2s3, Default({'subgroup': 'suboption1'}))
    with pytest.raises(KeyError, match="Group '/subgroup2' does not have child 'subgroup'"):
        _resolve_default(s2,   Default({'subgroup': 'suboption1'}))
    with pytest.raises(KeyError, match="Group '/subgroup' does not have child 'subgroup'"):
        _resolve_default(s1,   Default({'subgroup': 'suboption1'}))


def test_defaults_advanced():

    def resolve_entry_defaults(group):
        results = []
        for default in group.get_option('default').get_unresolved_defaults():
            results.append(_resolve_default(group, default))
        return results

    assert resolve_entry_defaults(_make_config_group(suboption='suboption1')) == [('/subgroup', ['/subgroup/suboption1'], ('subgroup',))]
    assert resolve_entry_defaults(_make_config_group(suboption='suboption2')) == [('/subgroup', ['/subgroup/suboption2'], ('subgroup',))]
    assert resolve_entry_defaults(_make_config_group(suboption=['suboption2'])) == [('/subgroup', ['/subgroup/suboption2'], ('subgroup',))]
    assert resolve_entry_defaults(_make_config_group(suboption=['suboption1', 'suboption2'])) == [('/subgroup', ['/subgroup/suboption1', '/subgroup/suboption2'], ('subgroup',))]

    assert resolve_entry_defaults(_make_config_group(suboption=None, suboption2='suboption1')) == [('/subgroup2/subgroup3', ['/subgroup2/subgroup3/suboption1'], ('subgroup2', 'subgroup3'))]
    assert resolve_entry_defaults(_make_config_group(suboption=None, suboption2='suboption2')) == [('/subgroup2/subgroup3', ['/subgroup2/subgroup3/suboption2'], ('subgroup2', 'subgroup3'))]




import re

from eunomia.config.nodes import SubNode

from tests.test_backend_obj import _make_config_group
from tests.util import temp_capture_stdout


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def test_debug_groups():
    root = _make_config_group(suboption='suboption1')

    with temp_capture_stdout() as out:
        root.debug_tree_print()
    color_out = out.getvalue()
    assert color_out == ' \x1b[90m\x1b[0m\x1b[35m/\x1b[0m\n \x1b[90m├\x1b[93m╌\x1b[0m \x1b[90m/:\x1b[0m \x1b[33mdefault\x1b[0m\n \x1b[90m├\x1b[95m─\x1b[0m \x1b[90m\x1b[0m\x1b[35m/subgroup\x1b[0m\n \x1b[90m│\x1b[0m  \x1b[90m├\x1b[93m╌\x1b[0m \x1b[90m/subgroup:\x1b[0m \x1b[33msuboption1\x1b[0m\n \x1b[90m│\x1b[0m  \x1b[90m╰\x1b[93m╌\x1b[0m \x1b[90m/subgroup:\x1b[0m \x1b[33msuboption2\x1b[0m\n \x1b[90m╰\x1b[95m─\x1b[0m \x1b[90m\x1b[0m\x1b[35m/subgroup2\x1b[0m\n    \x1b[90m╰\x1b[95m─\x1b[0m \x1b[90m/subgroup2\x1b[0m\x1b[35m/subgroup3\x1b[0m\n       \x1b[90m├\x1b[93m╌\x1b[0m \x1b[90m/subgroup2/subgroup3:\x1b[0m \x1b[33msuboption1\x1b[0m\n       \x1b[90m╰\x1b[93m╌\x1b[0m \x1b[90m/subgroup2/subgroup3:\x1b[0m \x1b[33msuboption2\x1b[0m\n'

    with temp_capture_stdout() as out:
        root.debug_tree_print(colors=False)
    normal_out = out.getvalue()
    assert normal_out == ' /\n ├╌ /: default\n ├─ /subgroup\n │  ├╌ /subgroup: suboption1\n │  ╰╌ /subgroup: suboption2\n ╰─ /subgroup2\n    ╰─ /subgroup2/subgroup3\n       ├╌ /subgroup2/subgroup3: suboption1\n       ╰╌ /subgroup2/subgroup3: suboption2\n'
    # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    assert normal_out == re.sub(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])', '', color_out)

    with temp_capture_stdout() as out:
        root.debug_tree_print(colors=False, show_defaults=True)
    normal_out_defaults = out.getvalue()
    assert normal_out_defaults == ' /\n ├╌ /: default [subgroup: suboption1]\n ├─ /subgroup\n │  ├╌ /subgroup: suboption1\n │  ╰╌ /subgroup: suboption2\n ╰─ /subgroup2\n    ╰─ /subgroup2/subgroup3\n       ├╌ /subgroup2/subgroup3: suboption1\n       ╰╌ /subgroup2/subgroup3: suboption2\n'

    root = _make_config_group(suboption=SubNode('suboption${=1}'))
    with temp_capture_stdout() as out:
        root.debug_tree_print(colors=False, show_defaults=True)
    normal_out_defaults_special = out.getvalue()
    assert normal_out_defaults_special == ' /\n ├╌ /: default [subgroup: suboption${=1}]\n ├─ /subgroup\n │  ├╌ /subgroup: suboption1\n │  ╰╌ /subgroup: suboption2\n ╰─ /subgroup2\n    ╰─ /subgroup2/subgroup3\n       ├╌ /subgroup2/subgroup3: suboption1\n       ╰╌ /subgroup2/subgroup3: suboption2\n'

    # TODO: test other flags and suboption cases


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from collections import defaultdict
from typing import Dict, Union, DefaultDict, List, Tuple

from eunomia.config import Group, Option
from eunomia.util._targets import _fn_get_module_path, _camel_to_snake, make_target_option


# ========================================================================= #
# Advanced Group                                                            #
# ========================================================================= #


class RegistryGroup(Group):

    """
    This group allows registration of functions or classes, constructing an
    option that has the _target_ parameter set, with default values obtained
    from the default parameters of the function.

    The option by default resides under the group that corresponds to the module
    that the function or class resides in.
    """

    def __init__(
            self,
            named_nodes: Dict[str, Union['Group', 'Option']] = None,
    ):
        super().__init__(named_nodes=named_nodes)
        # save all default options
        self._registered_all: DefaultDict[object, List[Option]] = defaultdict(list)
        self._registered_defaults: Dict[object, Tuple[bool, Option]] = {}

    def register_target(
            self,
            fn=None,
            # config settings
            name: str = None, path: str = None, is_default: bool = None,
            # target function
            target: str = None, params: dict = None, mode: str = 'any', keep_defaults: bool = True,
            # option params extras
            nest_path: str = None, data: dict = None, pkg: str = None, defaults: list = None
    ):
        def wrapper(func):
            self.register_target_fn(
                func,
                name=name, path=path, is_default=is_default,
                target=target, params=params, mode=mode, keep_defaults=keep_defaults,
                nest_path=nest_path, data=data, pkg=pkg, defaults=defaults,
            )
            return func
        # decorate correctly!
        if fn is not None:
            return wrapper(fn)
        else:
            return wrapper

    def register_target_fn(
            self,
            fn,
            # config settings
            name: str = None, path: str = None, is_default: bool = None,
            # target function
            target: str = None, params: dict = None, mode: str = 'any', keep_defaults: bool = True,
            # option params extras
            nest_path: str = None, data: dict = None, pkg: str = None, defaults: list = None
    ) -> 'Option':
        assert not self.has_parent, 'Can only register on the root node.'
        # make various defaults
        path = _fn_get_module_path(fn) if path is None else path
        name = _camel_to_snake(fn.__name__) if name is None else name
        # make option under the specified group
        group = self.get_group_recursive(path=path, make_missing=True)
        option = group.add_option(name, make_target_option(
            fn, target=target, params=params, mode=mode, keep_defaults=keep_defaults,
            nest_path=nest_path, data=data, pkg=pkg, defaults=defaults
        ))
        # register the option first!
        try:
            self._register_obj(fn, option, is_default)
        except Exception as e:
            group.del_option(name)
            raise e
        # done!
        return option

    def _register_obj(self, func, option, is_default=None):
        assert not self.has_parent, 'Can only register on the root node.'
        # register as a default
        if func in self._registered_defaults:
            explicit, _ = self._registered_defaults[func]
            if explicit:
                if is_default:
                    raise AssertionError(f'registered callable {func} for option: {option.keys} was previously explicitly registered as a default.')
            else:
                if is_default:
                    self._registered_defaults[func] = (True, option)
                elif is_default is None:
                    self._registered_defaults.setdefault(func, (False, option))
        else:
            if is_default or (is_default is None):
                self._registered_defaults.setdefault(func, (False if is_default is None else True, option))
        # keep track that this was registered
        self._registered_all[func].append(option)

    def get_registered_defaults(self, explicit_only=False) -> list:
        assert not self.has_parent, 'Can only register on the root node.'
        # get default options!
        defaults = {}
        for func, (explicit, option) in self._registered_defaults.items():
            if explicit_only and not explicit:
                continue
            k = option.abs_group_path
            if k in defaults:
                raise AssertionError(f'default for callable {func} corresponding to option: {option.keys} has already been added.')
            defaults[k] = option.key
        return [{k: v} for k, v in defaults.items()]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



import ast
import sys
from typing import Optional, Dict, Any
from asteval.astutils import UNSAFE_ATTRS, make_symbol_table, safe_mult, safe_add, safe_pow, safe_lshift


# ========================================================================= #
# Errors                                                                    #
# ========================================================================= #


class InterpreterError(Exception):
    """
    A general interpreter error.
    """


class UnsupportedLanguageFeatureError(InterpreterError):
    """
    If the python language feature is not supported.
    """


class DisabledLanguageFeatureError(UnsupportedLanguageFeatureError):
    """
    If the python language feature has been disabled
    """

# ========================================================================= #
# Interpreter                                                               #
# ========================================================================= #


class BaseInterpreter(object):

    def copy(self) -> 'BaseInterpreter':
        return BaseInterpreter()

    def _get_visit_name(self, node):
        return 'visit_' + node.__class__.__name__

    def _visit(self, node, *args, **kwargs):
        attr = self._get_visit_name(node)
        visitor = getattr(self, attr, self._visit_unknown)
        if visitor is None:
            visitor = self._visit_disabled
        result = visitor(node, *args, **kwargs)
        return result

    def _visit_disabled(self, node, *args, **kwargs):
        attr = self._get_visit_name(node)
        raise DisabledLanguageFeatureError(f'Language feature has been disabled: node={repr(node)}\n method: attr={repr(attr)} is None')

    def _visit_unknown(self, node, *args, **kwargs):
        attr = self._get_visit_name(node)
        raise UnsupportedLanguageFeatureError(f'Language feature not supported: node={repr(node)}\nCould not find method: attr={repr(attr)}')

    def interpret(self, node_or_string, strip_whitespace=True):
        """
        Based on ast.literal_eval & ast.NodeTransformer

        - (Safely?) evaluate an expression node or a string containing
          a Python expression.
        """
        if isinstance(node_or_string, str):
            if strip_whitespace:
                # string.whitespace == ' \t\n\r\v\f'
                # we dont want to strip special characters
                node_or_string = node_or_string.strip(' \t')
            # just future proof things in case... only supported for 3.8 and above
            if sys.version_info[:2] >= (3, 8):
                node_or_string = ast.parse(node_or_string, mode='eval', feature_version=(3, 7))
            else:
                node_or_string = ast.parse(node_or_string, mode='eval')
        if isinstance(node_or_string, ast.Expression):
            node_or_string = node_or_string.body
        return self._visit(node_or_string)


# ========================================================================= #
# Basic Interpreter for expressions and literals                            #
# ========================================================================= #


class BasicInterpreter(BaseInterpreter):
    """
    https://docs.python.org/3/reference/expressions.html
    https://docs.python.org/3/library/stdtypes.html
    """

    def __init__(
            self,
            # rules
            allow_nested_unary=False,
            allow_numerical_unary_on_bool=False,
            allow_chained_comparisons=True,
    ):
        # TODO: maybe convert to mixins?
        self._allow_nested_unary = allow_nested_unary
        self._allow_numerical_unary_on_bool = allow_numerical_unary_on_bool
        self._allow_chained_comparisons = allow_chained_comparisons

    def copy(self) -> 'BasicInterpreter':
        return BasicInterpreter(
            allow_nested_unary=self._allow_nested_unary,
            allow_numerical_unary_on_bool=self._allow_numerical_unary_on_bool,
            allow_chained_comparisons=self._allow_chained_comparisons,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Python 3.6 Support - Custom handling is required...                   #
    # these nodes do not exist in python 3.9, not sure about 3.8? 3.7?      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # TODO: deprecated! old python language feature!
    def visit_Str(self, node): return node.s
    # TODO: deprecated! old python language feature!
    def visit_Num(self, node): return node.n
    # TODO: deprecated! old python language feature!
    def visit_NameConstant(self, node): return node.value
    # TODO: deprecated! old python language feature!
    def visit_Bytes(self, node): return node.s

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Atoms ::= identifier | literal | enclosure                            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # literals ::= stringliteral | bytesliteral | integer | floatnumber | imagnumber
    def visit_Constant(self, node): return node.value

    # enclosures ::= parenth_form | list_display | dict_display | set_display | generator_expression | yield_atom
    def visit_List (self, node): return list(self._visit(n) for n in node.elts)
    def visit_Tuple(self, node): return tuple(self._visit(n) for n in node.elts)
    def visit_Set  (self, node): return set(self._visit(n) for n in node.elts)
    def visit_Dict (self, node):
        if len(node.keys) != len(node.values):
            raise ValueError('Dict node is malformed, differing number of keys and values!')
        return dict((self._visit(k), self._visit(v)) for k, v in zip(node.keys, node.values))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Boolean Operations                                                    #
    # docs.python.org/3/library/stdtypes.html#boolean-operations-and-or-not #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def visit_BoolOp(self, node):
        return self._visit(node.op, node.values)

    # 1. This is a short-circuit operator, so it only evaluates the second argument if the first one is true.
    def visit_Or(self, node, values):
        result = False
        for value in values:
            result = result or self._visit(value)
            if result:
                return True
        return False

    # 2. This is a short-circuit operator, so it only evaluates the second argument if the first one is false.
    def visit_And(self, node, values):
        result = True
        for value in values:
            result = result and self._visit(value)
            if not result:
                return False
        return True

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Comparisons                                                           #
    # docs.python.org/3/library/stdtypes.html#comparisons                   #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def visit_Compare(self, node):
        assert len(node.ops) == len(node.comparators)
        if self._allow_chained_comparisons:
            # Comparisons can be chained arbitrarily; for example, x < y <= z is
            # equivalent to x < y and y <= z, except that y is evaluated only once (but
            # in both cases z is not evaluated at all when x < y is found to be false)
            result = True
            vis_left = self._visit(node.left)
            for op, comp in zip(node.ops, node.comparators):
                vis_right = self._visit(comp)
                result = result and self._visit(op, vis_left, vis_right)
                if not result:
                    return False
                vis_left = vis_right
            return True
        else:
            if len(node.ops) > 1:
                raise DisabledLanguageFeatureError(f'Chained operators are not allowed, eg. (a <op> b <op> 3)\nConvert to ((a <op> b) and (b <op> c))')
            vis_left = self._visit(node.left)
            vis_right = self._visit(node.comparators[0])
            return self._visit(node.ops[0], vis_left, vis_right)

    def visit_Lt    (self, op, vis_left, vis_right): return vis_left < vis_right
    def visit_LtE   (self, op, vis_left, vis_right): return vis_left <= vis_right
    def visit_Gt    (self, op, vis_left, vis_right): return vis_left > vis_right
    def visit_GtE   (self, op, vis_left, vis_right): return vis_left >= vis_right
    def visit_Eq    (self, op, vis_left, vis_right): return vis_left == vis_right
    def visit_NotEq (self, op, vis_left, vis_right): return vis_left != vis_right
    def visit_Is    (self, op, vis_left, vis_right): return vis_left is vis_right
    def visit_IsNot (self, op, vis_left, vis_right): return vis_left is not vis_right

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Operators                                                             #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def visit_BinOp(self, node):
        return self._visit(node.op, self._visit(node.left), self._visit(node.right))

    def visit_Add     (self, node, vis_left, vis_right): return safe_add(vis_left, vis_right)
    def visit_Sub     (self, node, vis_left, vis_right): return vis_left - vis_right
    def visit_Mult    (self, node, vis_left, vis_right): return safe_mult(vis_left, vis_right)
    def visit_MatMult (self, node, vis_left, vis_right): return vis_left @ vis_right
    def visit_Div     (self, node, vis_left, vis_right): return vis_left / vis_right
    def visit_FloorDiv(self, node, vis_left, vis_right): return vis_left // vis_right
    def visit_Mod     (self, node, vis_left, vis_right): return vis_left % vis_right
    def visit_Pow     (self, node, vis_left, vis_right): return safe_pow(vis_left, vis_right)

    def visit_BitAnd  (self, node, vis_left, vis_right): return vis_left & vis_right
    def visit_BitOr   (self, node, vis_left, vis_right): return vis_left | vis_right
    def visit_BitXor  (self, node, vis_left, vis_right): return vis_left ^ vis_right

    def visit_LShift  (self, node, vis_left, vis_right): return safe_lshift(vis_left, vis_right)
    def visit_RShift  (self, node, vis_left, vis_right): return vis_left >> vis_right

    def visit_In   (self, node, vis_left, vis_right): return vis_left in vis_right
    def visit_NotIn(self, node, vis_left, vis_right): return vis_left not in vis_right

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Unary Operators                                                       #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def visit_UnaryOp(self, node):
        if not self._allow_nested_unary:
            if isinstance(node.operand, ast.UnaryOp):
                raise DisabledLanguageFeatureError(f'Nested unary operators are not allowed, eg. ++1 or --1')
        if not self._allow_numerical_unary_on_bool:
            if not isinstance(node.op, ast.Not):
                # TODO: deprecated language feature ast.NameConstant, not used in python 3.9
                if isinstance(node.operand, (ast.Constant, ast.NameConstant)) and isinstance(node.operand.value, bool):
                    raise DisabledLanguageFeatureError(f'Only the not unary operator is allowed on booleans, eg. not True')
        return self._visit(node.op, self._visit(node.operand))

    def visit_UAdd  (self, op, visited_value): return + visited_value
    def visit_USub  (self, op, visited_value): return - visited_value
    def visit_Invert(self, op, visited_value): return ~ visited_value
    def visit_Not   (self, op, visited_value): return not visited_value

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # F Strings                                                             #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def visit_JoinedStr(self, node):
        return ''.join(str(self._visit(v)) for v in node.values)

    def visit_FormattedValue(self, node):
        value = self._visit(node.value)
        if node.format_spec is None:
            return f'{value}'
        else:
            format = self._visit(node.format_spec)
            return f'{value:{format}}'


# ========================================================================= #
# Interpreter for properties, getters and symtables                         #
# ========================================================================= #


class Interpreter(BasicInterpreter):

    def __init__(
            self,
            # symtable
            default_symtable: Optional[Dict[str, Any]] = None,
            extra_symtable: Optional[Dict[str, Any]] = None,
            # rules
            allow_nested_unary=False,
            allow_numerical_unary_on_bool=False,
            allow_chained_comparisons=True,
            allow_unpacking=False,
            # NON-STANDARD-PYTHON
            NON_STANDARD_PYTHON_allow_getitem_on_getattr_fail=False,
    ):
        super().__init__(
            allow_nested_unary=allow_nested_unary,
            allow_numerical_unary_on_bool=allow_numerical_unary_on_bool,
            allow_chained_comparisons=allow_chained_comparisons,
        )
        self._allow_unpacking = allow_unpacking
        # make default symtable
        self._symtable = make_symbol_table(use_numpy=False) if (default_symtable is None) else default_symtable
        # add extras to symtable
        self._symtable.update({} if (extra_symtable is None) else extra_symtable)
        # THIS IS NON STANDARD PYTHON
        # THIS SHOULD MAYBE BE REPLACED WITH AN ATTR_DICT
        self._NON_STANDARD_PYTHON_allow_getitem_on_getattr_fail = NON_STANDARD_PYTHON_allow_getitem_on_getattr_fail

    def copy(self) -> 'Interpreter':
        # TODO: this is not safe, does not use a deep copy
        return Interpreter(
            default_symtable=dict(self._symtable),
            # rules
            allow_nested_unary=self._allow_nested_unary,
            allow_numerical_unary_on_bool=self._allow_numerical_unary_on_bool,
            allow_chained_comparisons=self._allow_chained_comparisons,
            allow_unpacking=self._allow_unpacking,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Properties                                                            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def visit_Name(self, node):
        try:
            return self._symtable[node.id]
        except KeyError:
            pass
        raise NameError(f"name {repr(node.id)} is not defined")

    def visit_Attribute(self, node):
        if node.attr in UNSAFE_ATTRS:
            raise KeyError(f'Tried to access unsafe attribute: {node.attr}')
        visited, attr = self._visit(node.value), node.attr
        try:
            return getattr(visited, attr)
        except AttributeError as e:
            if self._NON_STANDARD_PYTHON_allow_getitem_on_getattr_fail:
                if hasattr(visited, '__getitem__'):
                    try:
                        return visited[attr]
                    except KeyError:
                        pass
            raise e

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # List, Tuple, Set, Dict Unpacking                                      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _yield_visit_unpack_elts(self, elts) -> list:
        for n in elts:
            if isinstance(n, ast.Starred):
                if not self._allow_unpacking:
                    raise DisabledLanguageFeatureError('Starred unpacking is disabled.')
                yield from self._visit(n.value)
            else:
                yield self._visit(n)

    def _yield_visit_unpack_dict_pairs(self, keys, values, is_keywords=False) -> list:
        # same error as super().visit_Dict(...)
        if not is_keywords:
            if len(keys) != len(values):
                raise ValueError('Dict node is malformed, differing number of keys and values!')
        # actually unpack
        for k, v in zip(keys, values):
            if k is None:
                if not self._allow_unpacking:
                    raise DisabledLanguageFeatureError('Starred unpacking is disabled.')
                # get dictionary value
                unpack_dict = self._visit(v)
                # check that we are not malformed
                if not isinstance(unpack_dict, dict):
                    raise ValueError('Tried to unpack non dict.')
                # return all
                yield from unpack_dict.items()
            else:
                if not is_keywords:
                    yield self._visit(k), self._visit(v)
                else:
                    yield k, self._visit(v)

    def visit_List (self, node): return list(self._yield_visit_unpack_elts(node.elts))
    def visit_Tuple(self, node): return tuple(self._yield_visit_unpack_elts(node.elts))
    def visit_Set  (self, node): return set(self._yield_visit_unpack_elts(node.elts))
    def visit_Dict (self, node): return dict(self._yield_visit_unpack_dict_pairs(node.keys, node.values, is_keywords=False))

    def visit_Starred(self, node):
        raise RuntimeError('This should never happen.')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Call                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _build_kwargs(self, keywords: list):
        # unpack everything like a dictionary
        unpacked_keys_values = self._yield_visit_unpack_dict_pairs(
            keys=(kw.arg for kw in keywords),
            values=(kw.value for kw in keywords),
            is_keywords=True,
        )
        # check that kwargs are unique
        kwargs = {}
        for k, v in unpacked_keys_values:
            if k in kwargs:
                raise TypeError(f"got multiple values for keyword argument {repr(k)}")
            kwargs[k] = v
        # done!
        return kwargs

    def visit_Call(self, node):
        func = self._visit(node.func)
        # get args and kwargs
        args = self._yield_visit_unpack_elts(node.args) if node.args else []
        kwargs = self._build_kwargs(node.keywords) if node.keywords else {}
        # call the function
        return func(*args, **kwargs)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Get Item                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # TODO: maybe names, subscripts and properties should be moved into basic
    #       but with limitations on which funcs are available
    def visit_Subscript(self, node):
        slice = self._visit(node.slice)
        value = self._visit(node.value)
        return value[slice]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Python 3.6 Support - Custom handling is required...                   #
    # these nodes do not exist in python 3.9, not sure about 3.8? 3.7?      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # TODO: deprecated! old python language feature!
    def visit_Index(self, node):
        return self._visit(node.value)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # List Comprehension                                                    #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # def visit_ListComp(self, node):
    #
    #     # limitations - only one "for .. in"
    #     assert len(node.generators) == 1
    #     gen = node.generators[0]
    #     # limitations - no "if", no multiple assignment, no async
    #     assert len(gen.ifs) == 0
    #     assert not gen.is_async
    #     assert isinstance(gen.target, ast.Name)
    #
    #     # target name
    #     target: str = node.generators[0].target.id
    #
    #     # this is super slow
    #     results, interp_scope = [], self.copy()
    #     for value in self._visit(gen.iter):
    #         interp_scope._symtable[target] = value
    #         results.append(interp_scope._visit(node.elt))
    #
    #     return results

    # def test_interpreter_list_comprehension(name_sym_interp):
    #     # assert name_sym_interp('[i for i in [1, 2, 3, 4, 5]]') == [1, 2, 3, 4, 5]
    #     # assert name_sym_interp('[i for i, j in [1, 2, 3, 4, 5] if i == 4]') == [1, 2, 3, 4, 5]
    #     # assert name_sym_interp('[i - 1 for i in [1, 2, 3, 4, 5]]') == [0, 1, 2, 3, 4]
    #     # assert name_sym_interp('{i - 1 for i in [1, 2, 3, 4, 5]}') == {0, 1, 2, 3, 4}
    #     # assert name_sym_interp('{i: i - 1 for i in [1, 2, 3, 4, 5]}') == {1:0, 2:1, 3:2, 4:3, 5:4}
    #     # assert name_sym_interp('(i for i in [1, 2, 3, 4, 5])') == {1:0, 2:1, 3:2, 4:3, 5:4}
    #
    #     # assert name_sym_interp('[x+1 for x in [0,1,2,3,4]]') == [1, 2, 3, 4, 5]
    #     # assert name_sym_interp('[i for j in [[0],[1,2],[3,4]] for i in j]') == [0, 1, 2, 3, 4]
    #
    #     # [i for i in [1, 2, 3] if i for j in [3, 4, 5] if j if j]


def interpret_expr(
        string: str,
        usersyms: Optional[Dict[str, Any]] = None,
        NON_STANDARD_PYTHON = False,
):
    """
    Interpret the given expression with preset
    limitations on what is allowed.
    """
    interpreter = Interpreter(
        extra_symtable=usersyms,
        NON_STANDARD_PYTHON_allow_getitem_on_getattr_fail=NON_STANDARD_PYTHON
    )
    assert isinstance(string, str)
    return interpreter.interpret(string)



from numbers import Number
from typing import Iterable, List, Tuple, Type


# ========================================================================= #
# Recursive Dictionary Helper                                               #
# ========================================================================= #


def recursive_getitem(dct, keys: Iterable[str], make_missing=False):
    if not keys:
        return dct
    (key, *keys) = keys
    if make_missing:
        if key not in dct:
            dct[key] = {}
    return recursive_getitem(dct[key], keys, make_missing=make_missing)


def recursive_setitem(dct, keys: Iterable[str], value, make_missing=False):
    if not keys:
        raise KeyError(f'{recursive_setitem.__name__} requires as least one key: {keys}')
    (*keys, key) = keys
    insert_at = recursive_getitem(dct, keys, make_missing=make_missing)
    insert_at[key] = value


def dict_recursive_update(left, right, allow_overwrite=True, safe_merge=True, allow_update_types: List[Tuple[Type, ...]] = None):
    # check type groups
    if allow_update_types is None:
        allow_update_types = []
    # allow floats and integers to mix by default
    allow_update_types = [(Number,)] + allow_update_types
    # check user groups
    assert all(isinstance(group, tuple) for group in allow_update_types)
    # begin merge!
    _dict_recursive_update(left, right, [], allow_overwrite=allow_overwrite, safe_merge=safe_merge, type_merge_groups=allow_update_types)


def _dict_recursive_update(left, right, stack, allow_overwrite, safe_merge, type_merge_groups):
    # right overwrites left
    for k, rv in right.items():
        if k not in left:
            left[k] = rv
        else:
            # get values
            lv = left[k]
            # handle cases
            # -- l and r are dictionaries
            if isinstance(lv, dict) and isinstance(rv, dict):
                _dict_recursive_update(left[k], rv, stack=stack + [k], allow_overwrite=allow_overwrite, safe_merge=safe_merge, type_merge_groups=type_merge_groups)
                continue
            if not allow_overwrite:
                raise KeyError(f'cannot overwrite existing keys {stack}')
            # -- l and r have different types
            if type(lv) is not type(rv):
                if not any(isinstance(lv, group) and isinstance(rv, group) for group in type_merge_groups):
                    if safe_merge:
                        raise TypeError(f'with safe_merge=True, {dict_recursive_update.__name__} cannot update keys with different types l={type(lv).__name__} r={type(rv).__name__}.\nError occurred at key: {stack + [k]}')
            left[k] = rv

# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import inspect
import importlib
import re
from copy import deepcopy

from eunomia.util._util_dict import recursive_getitem, dict_recursive_update
from eunomia.util._util_traverse import RecursiveTransformer
from eunomia.config import keys as K
from eunomia.config import validate as V
from eunomia.config import Option


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _fn_get_kwargs(func) -> dict:
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def _fn_get_args(func) -> list:
    signature = inspect.signature(func)
    return [
        k
        for k, v in signature.parameters.items()
        if v.default is inspect.Parameter.empty
    ]


def _fn_get_all_args(func) -> list:
    signature = inspect.signature(func)
    return list(signature.parameters.keys())


def _fn_get_module_path(obj):
    # get package search paths
    import os
    import sys
    # get module path
    path = os.path.abspath(inspect.getmodule(obj).__file__)
    # return the shortest relative path from all the packages
    rel_paths = []
    for site in sys.path:
        site = os.path.abspath(site)
        if os.path.commonprefix([site, path]) == site:
            rel_paths.append(os.path.relpath(path, site))
    # get shortest rel path
    rel_paths = sorted(rel_paths, key=str.__len__)
    assert len(rel_paths) > 0, f'no valid path found to: {obj}'
    # get the correct path
    path = rel_paths[0]
    assert path.endswith('.py')
    path = path[:-len('.py')]
    return os.path.normpath(path)


def _fn_get_import_path(obj):
    module_path = '.'.join(_fn_get_module_path(obj).split('/'))
    return f'{module_path}.{obj.__name__}'


def _camel_to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


# ========================================================================= #
# Make Targets                                                              #
# ========================================================================= #


def make_target_option(
        fn,
        # target function
        target: str = None,
        params: dict = None,
        mode: str = 'any',
        keep_defaults: bool = True,
        # option params extras
        nest_path: str = None,
        data: dict = None,
        pkg: str = None,
        defaults: list = None,
) -> 'Option':
    # get various defaults
    data = {} if data is None else data

    # check nest path
    if nest_path is not None:
        nest_path, is_relative = V.split_package_path(nest_path)
        assert not is_relative, 'nest path must not be relative'
    else:
        nest_path = []

    # get the dictionary to merge the target into
    targ_merge_dict = recursive_getitem(data, nest_path, make_missing=True)
    if not isinstance(targ_merge_dict, dict):
        raise ValueError('nested object in data must be a dictionary that the target can be merged into.')
    if targ_merge_dict:
        raise ValueError('nested object in data must be empty, otherwise target conflicts can occur.')

    # make data for the option
    dict_recursive_update(
        left=targ_merge_dict,
        right=make_target_dict(fn, target=target, params=params, mode=mode, keep_defaults=keep_defaults),
        safe_merge=False,
    )

    # make option
    return Option(
        data=data,
        pkg=pkg,
        defaults=defaults
    )


def make_target_dict(
        func,
        # target function
        target: str = None,
        params: dict = None,
        mode: str = 'any',
        keep_defaults: bool = True,
) -> dict:
    # get default values
    target = _fn_get_import_path(func) if target is None else target
    overrides = {} if params is None else params

    # if we should include all the non-overridden default
    # parameters in the final config
    if keep_defaults:
        defaults = _fn_get_kwargs(func)
    else:
        defaults = {}

    # get the parameters that can be overridden
    args, kwargs, all_args = _fn_get_args(func), _fn_get_kwargs(func), _fn_get_all_args(func)
    if mode == 'kwargs':
        allowed_overrides = set(kwargs)
    elif mode == 'any':
        allowed_overrides = set(args) | set(kwargs)
    elif mode == 'all':
        allowed_overrides = set(args) | set(kwargs)
    elif mode == 'unchecked':
        allowed_overrides = None
    else:
        raise KeyError(f'Invalid override mode: {repr(mode)}')

    # generate final list of parameter overrides
    if mode == 'unchecked':
        defaults.update(overrides)
        overrides = {}
    else:
        for k in all_args:  # sorted
            if k in allowed_overrides:
                if k in overrides:
                    defaults[k] = overrides.pop(k)

    # check that no extra unused overrides exist
    # and that _target_ is not a parameter name
    assert not overrides, f'cannot override params: {list(overrides.keys())}'
    assert K.MARKER_KEY_TARGET not in defaults, f'object {target} has conflicting optional parameter: {K.MARKER_KEY_TARGET}'

    # check that nothing is left over according to the override mode
    if mode == 'all':
        missed_args = set(args) - set(defaults)
        if missed_args:
            missed_args = [a for a in all_args if a in missed_args]  # sorted
            raise AssertionError(f'all non-default parameters require an override: {missed_args}')

    # return final dictionary
    return {
        K.MARKER_KEY_TARGET: target,
        **defaults,
    }


# ========================================================================= #
# Instantiate Targets                                                       #
# ========================================================================= #


def import_module_attr(target):
    # check the keys
    try:
        keys = V.validate_identifier_list(str.split(target, '.'))
    except Exception as e:
        raise ValueError(f'target {repr(target)} must contain at least one module component, and one attribute component eg: "math.log" or one builtin eg: "dict". Given target is: {repr(target)}').with_traceback(e.__traceback__)
    # special case, handle all builtins like dict or list
    if len(keys) < 1: # pragma: no cover
        raise RuntimeError('This is a bug and should not happen!')
    if len(keys) < 2:
        keys = ['builtins'] + keys
    # [1, len) : leaves at least one module keys, and at least one attr keys. Never empty
    for i in reversed(range(1, len(keys))):
        module_keys, attr_keys = keys[:i], keys[i:]
        # get the module, importlib cant import attributes on modules.
        try:
            module = importlib.import_module('.'.join(module_keys))
        except:
            continue
        # get the attr on the module
        obj = module
        try:
            for attr in attr_keys:
                obj = getattr(obj, attr)
        except AttributeError as e:
            raise AttributeError(f'could not import target {repr(target)}. {e}')
        return obj
    # we failed...
    raise ImportError(f'could not import target {repr(target)}. Are you sure the import module is correct?')


class _InstantiateTransformer(RecursiveTransformer):

    def __init__(self, recursive: bool = True):
        self._recursive = recursive
        self._visited = False

    def transform(self, value):
        # if not recursive, only go one layer deep
        if not self._recursive:
            if self._visited:
                return value
            self._visited = True
        # recurse like usual!
        return super().transform(value)

    def __transform_default__(self, value):
        # we dont throw an error if we encounter an unknown type
        return value

    def _transform_dict(self, config):
        # transform like usual
        config = super()._transform_dict(config)
        # return early
        if K.MARKER_KEY_TARGET not in config:
            return config
        # get args and kwargs
        target = config.pop(K.MARKER_KEY_TARGET)
        args = config.pop(K.MARKER_KEY_ARGS, ())
        kwargs = config.pop(K.MARKER_KEY_KWARGS, config)
        # if kwargs key existed in the config
        # check that config is empty after popping everything
        if (kwargs is not config) and config:
            raise KeyError(f'target dictionary contains the {K.MARKER_KEY_KWARGS} key as well as unused keys, which otherwise would be extracted as the kwargs. Choose one mode or the other.')
        # import & call
        fn = import_module_attr(target)
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            raise RuntimeError(f'Could not call target {repr(target)} object {fn} with args: {args} and kwargs: {kwargs}').with_traceback(e.__traceback__)


def instantiate(config, recursive=True, ensure_root_target=True):
    # if the root must be a dictionary and valid target
    if ensure_root_target:
        if not isinstance(config, dict):
            raise TypeError(f'Could not instantiate root object: not a dictionary! Otherwise set ensure_root_target=False')
        if K.MARKER_KEY_TARGET not in config:
            raise KeyError(f'Could not instantiate root object: target marker key {repr(K.MARKER_KEY_TARGET)} not found in root! Otherwise set ensure_root_target=False')
    # WARNING: this can mutate the given config
    config = deepcopy(config)
    # transform!
    return _InstantiateTransformer(recursive=recursive).transform(config)


call = instantiate


# ========================================================================= #
# END                                                                       #
# ========================================================================= #




from pathlib import Path
from typing import Union

from eunomia.backend import BackendYaml, BackendDict, BackendObj, Backend
from eunomia.config import Group


DefaultConfigTypes = Union[str, Path, dict, Group]


def infer_backend_load_group(config: DefaultConfigTypes, backend: Backend = None) -> Group:
    # override backend
    if backend is not None:
        return backend.load_group(config)
    # infer backend
    if isinstance(config, str):
        return BackendYaml().load_group(config)
    elif isinstance(config, dict):
        # we assume that the config is a dictionary
        return BackendDict().load_group(config)
    elif isinstance(config, Group):
        # we assume that the config is a ConfigGroup with ConfigOptions
        return BackendObj().load_group(config)
    else:
        raise TypeError(f'Unsupported config_root type: {config.__class__.__name__}')



import sys
from typing import Callable, Any
import pytest
import ast
from eunomia.config.nodes._util_interpret import Interpreter, BasicInterpreter
from eunomia.config.nodes._util_interpret import DisabledLanguageFeatureError


# ========================================================================= #
# Fixtures                                                                  #
# ========================================================================= #


@pytest.fixture()
def interpret() -> Callable[[str], Any]:
    return BasicInterpreter().interpret


@pytest.fixture()
def interpret_nonstrict() -> Callable[[str], Any]:
    return BasicInterpreter(
        allow_nested_unary=True,
        allow_numerical_unary_on_bool=True,
        allow_chained_comparisons=True,
    ).interpret


# ========================================================================= #
# Basic Expressions + Literals                                              #
# ========================================================================= #


# https://docs.python.org/3/reference/expressions.html#literals
def test_interpret_literals(interpret):
    # stringliteral | bytesliteral
    assert interpret('"asdf"') == "asdf"
    assert interpret('"asdf" \'fdsa\'') == "asdffdsa"
    assert interpret('b"asdf"') == b"asdf"
    assert interpret('b"asdf" b\'fdsa\'') == b"asdffdsa"
    with pytest.raises(SyntaxError):
        interpret('b"asdf" "fdsa"')
    with pytest.raises(SyntaxError):
        interpret('"asdf" b"fdsa"')
    # integer
    assert interpret('1') == 1
    assert interpret('1_000') == 1_000
    # floatnumber
    assert interpret('1.0') == 1.0
    assert interpret('1.') == 1.
    assert interpret('.1') == .1
    assert interpret('1e-3') == 1e-3
    # imagnumber
    assert interpret('1j') == 1j
    assert interpret('-1j') == -1j
    assert interpret('1 - 1j') == 1 - 1j
    assert interpret('1 + 1j') == 1 + 1j


# https://docs.python.org/3/reference/expressions.html#displays-for-lists-sets-and-dictionaries
def test_interpret_lists_sets_dicts_tuples(interpret):
    # lists
    # assert interpret('list()') == list()
    assert interpret('[]') == []
    assert interpret('[1, 2, 3]') == [1, 2, 3]
    # sets
    # assert interpret('set()') == set()
    assert interpret('{1, 2, 3, 3}') == {1, 2, 3, 3}
    # dicts
    # assert interpret('dict()') == dict()
    assert interpret('{}') == dict()
    assert interpret('{}') != set()
    assert interpret('{1: 11, 2: 22, 3: 33, 3: 44}') == {1: 11, 2: 22, 3: 33, 3: 44}
    # tuples
    # assert interpret('tuple()') == tuple()
    assert interpret('()') == ()
    assert interpret('(1)') != (1,)
    assert interpret('(1,)') == (1,)
    assert interpret('(1, 2, 3)') == (1, 2, 3)
    assert interpret('(1, 2, 3)') != [1, 2, 3]
    # combined
    assert interpret('([1, (2, 3)], {1, 2}, {1: 11, 2: 22})') == ([1, (2, 3)], {1, 2}, {1: 11, 2: 22})


def test_interpret_starred_lists_sets_dicts_tuples(interpret):
    # TODO: assert interpret('(1, *[2, 3], 4)') == (1, *[2, 3], 4)
    # TODO: assert interpret('[1, *(2, 3), 4]') == [1, *(2, 3), 4]
    # TODO: assert interpret('{**{1: 11, 3: 33}, **{2: 22, 3: 44}}') == {**{1: 11, 3: 33}, **{2: 22, 3: 44}}
    pass


# https://docs.python.org/3/library/stdtypes.html#boolean-operations-and-or-not
def test_interpret_boolean_operations(interpret):
    assert interpret('not True') == (not True)
    # test short circuit
    assert interpret('False or False or True') == (False or False or True)
    assert interpret('True and True and False') == (True and True and False)
    # test not short circuit
    assert interpret('False or False or False') == (False or False or False)
    assert interpret('True and True and True') == (True and True and True)


def test_interpret_binary_operators(interpret):
    assert interpret('9 + 2')  == 9 + 2
    assert interpret('9 - 2')  == 9 - 2
    assert interpret('9 * 2')  == 9 * 2
    # assert interpret('9 @ 2')  == 9 @ 2
    assert interpret('9 / 2')  == 9 / 2
    assert interpret('9 // 2') == 9 // 2
    assert interpret('9 % 2')  == 9 % 2
    assert interpret('9 ** 2') == 9 ** 2
    assert interpret('9 & 2')  == 9 & 2
    assert interpret('9 | 2')  == 9 | 2
    assert interpret('9 ^ 2')  == 9 ^ 2
    assert interpret('9 << 2') == 9 << 2
    assert interpret('9 >> 2') == 9 >> 2


def test_interpret_comparisons(interpret):
    assert interpret('9 < 2') == (9 < 2)
    assert interpret('9 <= 2') == (9 <= 2)
    assert interpret('9 > 2') == (9 > 2)
    assert interpret('9 >= 2') == (9 >= 2)
    assert interpret('9 == 2') == (9 == 2)
    assert interpret('9 != 2') == (9 != 2)
    assert interpret('None is None') == (None is None)
    assert interpret('None is not None') == (None is not None)


def test_interpret_comparisons_chained(interpret, interpret_nonstrict):
    with pytest.raises(DisabledLanguageFeatureError):
        BasicInterpreter(allow_chained_comparisons=False).interpret('1 <= 2 <= 4')
    assert interpret('1 <= 2 <= 4') == (1 <= 2 <= 4)
    assert interpret_nonstrict('1 <= 2 <= 4') == (1 <= 2 <= 4)
    assert interpret_nonstrict('1 <= 4 <= 2') == (1 <= 4 <= 2)


def test_interpret_unary_ops(interpret):
    # https://docs.python.org/3/library/stdtypes.html#numeric-types-int-float-complex
    assert interpret('+1') == +1
    assert interpret('-1') == -1
    assert interpret('+1j') == +1j
    assert interpret('-1j') == -1j
    # https://docs.python.org/3/library/stdtypes.html#boolean-operations-and-or-not
    assert interpret('not True') == (not True)
    # https://docs.python.org/3/library/stdtypes.html#bitwise-operations-on-integer-types
    assert interpret('~1') == ~1


def test_interpret_fstrings(interpret):
    assert interpret('f"string"') == "string"
    assert interpret('f"prefix{1}suffix"') == f"prefix{1}suffix"
    assert interpret('f"prefix{1}suffix{2}"') == f"prefix{1}suffix{2}"
    assert interpret('f"{1}"') == f"{1}"
    assert interpret('f"{1}{2}"') == f"{1}{2}"
    with pytest.raises(SyntaxError):
        assert interpret('f"{}"')


def test_interpret_fstrings_formatting(interpret):
    assert interpret('f"{1.012612313:1.3f}"') == "1.013"
    assert interpret('f"{[1,2,3]}"') == "[1, 2, 3]"
    assert interpret('f"{[1,{1,2,2},3]}"') == "[1, {1, 2}, 3]"
    assert interpret('f"{\'asdf\'}"') == "asdf"
    assert interpret('f"{1 + 2.56:1.1f}"') == "3.6"
    assert interpret('f"{1 + 2.54:1.1f}"') == "3.5"
    assert interpret('f"{1 + 2.546:1.{1+1}f}"') == "3.55"


# ========================================================================= #
# Interpreter                                                               #
# ========================================================================= #


def test_custom_interpreter(interpret):
    # make custom interpreter
    class CustomInterpreter(Interpreter):
        visit_Dict = None
    c_interpr = CustomInterpreter().interpret

    # specific feature disabled
    assert interpret('({1: 11}, 1)') == ({1: 11}, 1)
    with pytest.raises(DisabledLanguageFeatureError):
        assert c_interpr('({1: 11}, 1)')

    # something else still works
    assert interpret('({1, 2}, 1)') == ({1, 2}, 1)
    assert c_interpr('({1, 2}, 1)') == ({1, 2}, 1)


# TODO: these are not supported
# def test_interp_init_types(interpret):
#     assert interpret('list()') == list()
#     assert interpret('set()') == set()
#     assert interpret('dict()') == dict()
#     assert interpret('tuple()') == tuple()

# ========================================================================= #
# Original ast.literal_eval tests                                           #
# Relavent copyright applies for this section, I am not the owner.          #
# ========================================================================= #


def test_literal_eval(interpret, interpret_nonstrict):
    assert interpret('[1, 2, 3]') == [1, 2, 3]
    assert interpret('{"foo": 42}') == {"foo": 42}
    assert interpret('(True, False, None)') == (True, False, None)
    assert interpret('{1, 2, 3}') == {1, 2, 3}
    assert interpret('b"hi"') == b"hi"
    # assert interpret('set()') == set()
    assert interpret('6') == 6
    assert interpret('+6') == 6
    assert interpret('-6') == -6
    assert interpret('3.25') == 3.25
    assert interpret('+3.25') == 3.25
    assert interpret('-3.25') == -3.25
    assert repr(interpret('-0.0')) == '-0.0'
    # This is technically valid python... allow it
    with pytest.raises(DisabledLanguageFeatureError):
        assert interpret('++6') == 6
    assert interpret_nonstrict('++6') == 6
    # avoid SyntaxWarning on code coverage
    unary_true = +True
    with pytest.raises(DisabledLanguageFeatureError):
        assert interpret('+True') is unary_true
    assert interpret_nonstrict('+True') is unary_true
    # We want this...
    # with pytest.raises(UnsupportedLanguageFeatureError):
    assert interpret('2+3') == 5


def test_literal_eval_complex(interpret):
    # Issue #4907
    assert interpret('6j') == 6j
    assert interpret('-6j') == -6j
    assert interpret('6.75j') == 6.75j
    assert interpret('-6.75j') == -6.75j
    assert interpret('3+6j') == 3 + 6j
    assert interpret('-3+6j') == -3 + 6j
    assert interpret('3-6j') == 3 - 6j
    assert interpret('-3-6j') == -3 - 6j
    assert interpret('3.25+6.75j') == 3.25 + 6.75j
    assert interpret('-3.25+6.75j') == -3.25 + 6.75j
    assert interpret('3.25-6.75j') == 3.25 - 6.75j
    assert interpret('-3.25-6.75j') == -3.25 - 6.75j
    assert interpret('(3+6j)') == 3 + 6j
    # this is valid though... allow it
    # with pytest.raises(ValueError):
    interpret('-6j+3')
    # with pytest.raises(ValueError):
    interpret('-6j+3j')
    # with pytest.raises(ValueError):
    interpret('3+-6j')
    # with pytest.raises(ValueError):
    interpret('3+(0+6j)')
    # with pytest.raises(ValueError):
    interpret('-(3+6j)')


def test_literal_eval_malformed_dict_nodes(interpret):
    malformed = ast.Dict(keys=[ast.Constant(1), ast.Constant(2)], values=[ast.Constant(3)])
    with pytest.raises(ValueError):
        interpret(malformed)
    malformed = ast.Dict(keys=[ast.Constant(1)], values=[ast.Constant(2), ast.Constant(3)])
    with pytest.raises(ValueError):
        interpret(malformed)


def test_literal_eval_trailing_ws(interpret):
    assert interpret("    -1") == -1
    assert interpret("\t\t-1") == -1
    assert interpret(" \t -1") == -1
    with pytest.raises(IndentationError):
        interpret("\n -1")
    # extra
    assert interpret(" \t -1 \t ") == -1


def test_literal_eval_malformed_lineno(interpret, interpret_nonstrict):
    # but this stuff is valid... we want to allow it
    with pytest.raises(DisabledLanguageFeatureError, match=r'Nested unary operators are not allowed'):
        interpret("{'a': 1,\n'b':2,\n'c':++3,\n'd':4}")
    interpret_nonstrict("{'a': 1,\n'b':2,\n'c':++3,\n'd':4}")
    node = ast.UnaryOp(ast.UAdd(), ast.UnaryOp(ast.UAdd(), ast.Constant(6)))
    assert getattr(node, 'lineno', None) is None
    with pytest.raises(DisabledLanguageFeatureError, match=r'Nested unary operators are not allowed'):
        interpret(node)
    interpret_nonstrict(node)


# ========================================================================= #
# Original ast.literal_eval tests                                           #
# ========================================================================= #



class name:
    foo = 1
    class bar:
        baz = 2
    @staticmethod
    def taz(a=100, b=200):
        return 10 + a + b


conf = {'a': {'c': 1}, 'b': 2}


@pytest.fixture()
def name_sym_interp():
    interpret = Interpreter(
        default_symtable={'name': name, 'range': range, 'conf': conf},
        allow_unpacking=True,
    ).interpret
    return interpret


def test_interpreter_names(name_sym_interp):
    assert name_sym_interp('name') is name
    with pytest.raises(NameError):
        assert name_sym_interp('_unknown') is name.bar


def test_interpreter_properties(name_sym_interp):
    assert name_sym_interp('name.foo') is name.foo
    assert name_sym_interp('name.bar') is name.bar
    with pytest.raises(AttributeError):
        assert name_sym_interp('name._unknown') is name.bar

def test_interpreter_unpack(name_sym_interp):
    assert name_sym_interp('{*"abcaaaa"}') == {"a", "b", "c"} == {*"abcaaaa"}
    assert name_sym_interp('{*[1, 2], *(1, 2, 3)}') == {1, 2, 3}
    assert name_sym_interp('{1: 1.0, **{2: 2.0}, 3: 3.0}') == {1: 1.0, 2: 2.0, 3: 3.0}
    assert name_sym_interp('{1: 1.0, **{2: 2.0, **{4: 4.0}, **{4: 5.0}}, 3: 3.0}') == {1: 1.0, 2: 2.0, 3: 3.0, 4: 5.0}
    assert name_sym_interp('[1, *[2, *(3, 4), *{1, 2, 2}, 5], *{6: 6., 7: 7.}]') == [1, 2, 3, 4, 1, 2, 5, 6, 7]
    with pytest.raises(SyntaxError, match='invalid syntax'):
        assert name_sym_interp('[1, *[2, *(3, 4), *{1, 2, 2}, 5], **{6: 6., 7: 7.}]')

def test_interpreter_call(name_sym_interp):
    assert name_sym_interp('range(5)') == range(5)
    assert name_sym_interp('name.taz(1)') == name.taz(1)
    assert name_sym_interp('name.taz(a=1)') == name.taz(a=1)
    assert name_sym_interp('name.taz(b=2)') == name.taz(b=2)
    assert name_sym_interp('name.taz(a=1, b=2)') == name.taz(a=1, b=2)
    assert name_sym_interp('name.taz(1, b=2)') == name.taz(1, b=2)
    # technically this should raise SyntaxError, only TypeError for kwargs
    # this only works for python 3.9
    if sys.version_info[:2] >= (3, 9):
        with pytest.raises(TypeError, match="got multiple values for keyword argument 'b'"):
            name_sym_interp('name.taz(b=1, b=2)')
    else:
        with pytest.raises(SyntaxError, match="keyword argument repeated"):
            name_sym_interp('name.taz(b=1, b=2)')

def test_interpreter_call_and_unpack(name_sym_interp):
    # CALL TESTS
    assert name_sym_interp('[*range(5)]') == [0, 1, 2, 3, 4]

    # CALL & UNPACK TESTS
    assert name_sym_interp('name.taz(**{"b": 2})') == name.taz(**{"b": 2})
    assert name_sym_interp('name.taz(*[1], **{"b": 2})') == name.taz(*[1], **{"b": 2})
    assert name_sym_interp('name.taz(**{**{"b": 2}, **{"b": 3}})') == name.taz(b=3)
    with pytest.raises(TypeError, match="got multiple values for keyword argument 'a'"):
        name_sym_interp('name.taz(*[111], **{"a": 1}, **{"a": 11}, **{"b": 2})')


def test_interpreter_getitem(name_sym_interp):
    assert name_sym_interp('conf') is conf
    assert name_sym_interp('conf["a"]') == conf["a"]
    assert name_sym_interp('conf["b"]') == conf["b"]
    assert name_sym_interp('conf["a"]["c"]') == conf["a"]["c"]


def test_interpreter_NON_STANDARD_PYTHON():
    config = {'foo': 1, 'bar': {'baz': 2}}
    standard = Interpreter(extra_symtable={'conf': config})
    non_standard = Interpreter(extra_symtable={'conf': config}, NON_STANDARD_PYTHON_allow_getitem_on_getattr_fail=True)

    # check standard - get_item
    assert standard.interpret('conf["foo"]') == 1
    assert standard.interpret('conf["bar"]["baz"]') == 2
    with pytest.raises(AttributeError): standard.interpret('conf.foo')
    with pytest.raises(AttributeError): standard.interpret('conf.bar.baz')
    with pytest.raises(AttributeError): standard.interpret('conf["bar"].baz')
    with pytest.raises(AttributeError): standard.interpret('conf.bar["baz"]')
    # check standard - missing
    with pytest.raises(KeyError): standard.interpret('conf["buzz"]')
    with pytest.raises(KeyError): standard.interpret('conf["bar"]["buzz"]')
    with pytest.raises(AttributeError): standard.interpret('conf["bar"].buzz')
    with pytest.raises(AttributeError): standard.interpret('conf.bar["buzz"]')

    # check standard - get_item
    assert non_standard.interpret('conf["foo"]') == 1
    assert non_standard.interpret('conf["bar"]["baz"]') == 2
    assert non_standard.interpret('conf.foo') == 1
    assert non_standard.interpret('conf.bar.baz') == 2
    assert non_standard.interpret('conf["bar"].baz') == 2
    assert non_standard.interpret('conf.bar["baz"]') == 2
    # check standard - missing
    with pytest.raises(KeyError): non_standard.interpret('conf["buzz"]')
    with pytest.raises(KeyError): non_standard.interpret('conf["bar"]["buzz"]')
    with pytest.raises(AttributeError): non_standard.interpret('conf["bar"].buzz')  # different from above
    with pytest.raises(KeyError): non_standard.interpret('conf.bar["buzz"]')

# def test_interpreter_list_comprehension(name_sym_interp):
#     # assert name_sym_interp('[i for i in [1, 2, 3, 4, 5]]') == [1, 2, 3, 4, 5]
#     # assert name_sym_interp('[i for i, j in [1, 2, 3, 4, 5] if i == 4]') == [1, 2, 3, 4, 5]
#     # assert name_sym_interp('[i - 1 for i in [1, 2, 3, 4, 5]]') == [0, 1, 2, 3, 4]
#     # assert name_sym_interp('{i - 1 for i in [1, 2, 3, 4, 5]}') == {0, 1, 2, 3, 4}
#     # assert name_sym_interp('{i: i - 1 for i in [1, 2, 3, 4, 5]}') == {1:0, 2:1, 3:2, 4:3, 5:4}
#     # assert name_sym_interp('(i for i in [1, 2, 3, 4, 5])') == {1:0, 2:1, 3:2, 4:3, 5:4}
#
#     # assert name_sym_interp('[x+1 for x in [0,1,2,3,4]]') == [1, 2, 3, 4, 5]
#     # assert name_sym_interp('[i for j in [[0],[1,2],[3,4]] for i in j]') == [0, 1, 2, 3, 4]
#
#     # [i for i in [1, 2, 3] if i for j in [3, 4, 5] if j if j]


# ================= #
# OLD ASTEVAL TESTS #
# ================= #

# import pytest
# from collections import namedtuple
# from eunomia._eval import try_parse_expression, asteval_interpret_expr
# from eunomia._eval import ParserError, ExpressionSemicolonError, UnsupportedLanguageFeatureError, NotExpressionError, MalformedPythonError
#
#
# def test_parse():
#     try_parse_expression('1')
#     try_parse_expression('1 + 2')
#     try_parse_expression('"asdf" + "fdsa"')
#     try_parse_expression('f"as{1}df" + "fdsa"')
#     try_parse_expression('config["asdf"]["fdsa"]["cheese"]')
#     try_parse_expression('config.asdf.fdsa')
#     try_parse_expression('[i for i in range(10)]')
#
#     with pytest.raises(ExpressionSemicolonError):
#         try_parse_expression('[1 + 2];')
#     with pytest.raises(NotExpressionError):
#         try_parse_expression('[1 + 2];')
#     with pytest.raises(ParserError):
#         try_parse_expression('[1 + 2];')
#
#     with pytest.raises(MalformedPythonError):
#         try_parse_expression('def asdf():\n    "full indent"\n  "partial indent"')
#     with pytest.raises(MalformedPythonError):
#         try_parse_expression('$2341(')
#     with pytest.raises(ParserError):
#         try_parse_expression('$2341(')
#
#     try_parse_expression('1 + 2')
#     with pytest.raises(NotExpressionError):
#         try_parse_expression('1 + 2;')
#     with pytest.raises(ParserError):
#         try_parse_expression('1 + 2;')
#
#     with pytest.raises(NotExpressionError):
#         try_parse_expression('a = [1 + 2]')
#     with pytest.raises(NotExpressionError):
#         try_parse_expression('a = [1 + 2]; b=2')
#     with pytest.raises(NotExpressionError):
#         try_parse_expression('a = [1 + 2]\nb=2')
#     with pytest.raises(NotExpressionError):
#         try_parse_expression('def func(): 1 + 2')
#
#     # these are not supported by asteval but should be
#     # supported here because they are expressions.
#     try_parse_expression('lambda: 1 + 2')
#     try_parse_expression('lambda: 1 + 2')
#     try_parse_expression('f"as{1}ds"')
#     try_parse_expression('f"as{1}ds" + "fdsa"')
#
#
# def test_eval():
#     assert asteval_interpret_expr('1') == 1
#     assert asteval_interpret_expr('1 + 2') == 3
#     assert asteval_interpret_expr('"asdf" + "fdsa"') == 'asdffdsa'
#     assert asteval_interpret_expr('[i for i in range(3)]') == [0, 1, 2]
#     assert asteval_interpret_expr('tuple([i for i in range(3)])') == (0, 1, 2)
#
#     with pytest.raises(UnsupportedLanguageFeatureError):
#         asteval_interpret_expr('f"as{1}df" + "fdsa"')
#     with pytest.raises(UnsupportedLanguageFeatureError):
#         asteval_interpret_expr('lambda: 1 + 2')
#     with pytest.raises(ExpressionSemicolonError):
#         asteval_interpret_expr('1 + 2;')
#     with pytest.raises(NotExpressionError):
#         asteval_interpret_expr('a = 1 + 2')
#     with pytest.raises(MalformedPythonError):
#         asteval_interpret_expr('%adsf@')
#
#     assert asteval_interpret_expr('config["key"][1]', usersyms=dict(config={'key': [1, 2, 3]})) == 2
#     assert asteval_interpret_expr('config.asdf["key"]', usersyms=dict(config=namedtuple('Test', 'asdf')(asdf={'key': 2}))) == 2
#
#
# def test_interpret_timeit():
#     def ave_timeit_ms(expr, n=100):
#         return f'{timeit(expr, number=n) / n * 1000:1.3f}ms'
#     expr = 'tuple([i for i in range(3)])'
#     def check_expr():
#         try_parse_expression(expr)
#     def check_invalid_expr():
#         try:
#             try_parse_expression(expr + ';')
#         except:
#             pass
#     def interpret_expr():
#         assert asteval_interpret_expr(expr) == (0, 1, 2)
#     print()
#     print('Check Expr Time:', ave_timeit_ms(check_expr))
#     print('Check Invalid Expr Time:', ave_timeit_ms(check_invalid_expr))
#     print('Interpret Time:', ave_timeit_ms(interpret_expr))


import pytest
from eunomia.backend._backend_yaml import yaml_load
from eunomia.config.nodes import IgnoreNode


# ========================================================================= #
# Test YAML & Custom Tags                                                   #
# ========================================================================= #


def test_yaml_tags():
    assert yaml_load('1') == 1
    assert yaml_load('[1, 2, 3]') == [1, 2, 3]
    assert yaml_load('{1, 2, 3}') == {1: None, 2: None, 3: None}

    # string tag - only compatible with scalars
    assert yaml_load('!str f""') == IgnoreNode('f""')
    assert yaml_load('!str 1') == IgnoreNode('1')
    with pytest.raises(TypeError):
        assert yaml_load('!str {1, 2, 3}') == IgnoreNode('{1: None, 2: None, 3: None}')
    with pytest.raises(TypeError):
        assert yaml_load('!str [1, 2, 3]') == '[1, 2, 3]'



from typing import Dict, Union, List, Tuple

from eunomia.util._util_traverse import RecursiveTransformer
from eunomia.config.nodes import ConfigNode, SubNode

from eunomia.config import validate as V
from eunomia.config import keys as K


# ========================================================================= #
# Config Node                                                               #
# ========================================================================= #


class _ConfigObject(object):

    def __init__(self):
        super().__init__()
        self._parent = None
        self._key = None
        self._children = {}

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Path                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def _is_valid_parent(self):
        return not ((not self._key) ^ (not self._parent))

    @property
    def has_parent(self):
        assert self._is_valid_parent
        return (self._key is not None) and (self._parent is not None)

    @property
    def parent(self) -> '_ConfigObject':
        assert self._is_valid_parent
        return self._parent

    @property
    def key(self) -> str:
        assert self._is_valid_parent
        return self._key

    @property
    def keys(self) -> Tuple[str]:
        return tuple(n.key for n in self.walk_from_root(visit_root=False))

    def get_package_keys(self, pkg_override: str = None) -> Tuple[str]:
        raise NotImplementedError

    @property
    def group(self) -> 'Group':
        raise NotImplementedError

    @property
    def abs_path(self):
        return V.keys_as_abs_config_path(self.keys)

    @property
    def root(self) -> '_ConfigObject':
        node = self
        for node in self.walk_to_root(visit_root=True):
            pass
        return node

    def rel_keys_to(self, right: '_ConfigObject'):
        if right.root is not self.root:
            raise AssertionError('trying to get relative path to config object in different tree.')
        # check that we are a parent
        l_keys, r_keys = self.keys, right.keys
        while l_keys:
            if not r_keys:
                raise KeyError('left config object is a child of the right config object')
            (l_key, *l_keys), (r_key, *r_keys) = l_keys, r_keys
            if l_key != r_key:
                raise KeyError('left config object is not a parent of the right config object')
        return r_keys

    def rel_path_to(self, right: '_ConfigObject'):
        return '/'.join(self.rel_keys_to(right))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Walk                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def walk_to_root(self, visit_root=False):
        current = self
        while current.has_parent:
            yield current
            current = current.parent
        if visit_root:
            yield current

    def walk_from_root(self, visit_root=False):
        yield from reversed(list(self.walk_to_root(
            visit_root=visit_root
        )))

    def walk_descendants(self):
        def _iter(node):
            yield node
            for k in node:
                yield from _iter(node[k])
        return _iter(self)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Children                                                              #
    # - for simplicity we do not allow attribute access to children.        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def add_child(self, key: str, child: '_ConfigObject') -> '_ConfigObject':
        if not isinstance(child, _ConfigObject):
            raise TypeError(f'child must be an instance of {_ConfigObject.__name__}')
        if child.has_parent:
            raise ValueError(f'child already has a parent, and cannot be added.')
        if key in self._children:
            raise KeyError(f'parent already has child with key: {key}')
        # set details
        child._parent = self
        child._key = key
        self._children[key] = child
        # return the added value
        return child

    def del_child(self, key: str):
        child = self.get_child(key)
        # remove details
        child._parent = None
        child._key = None
        del self._children[key]
        return child

    def get_child(self, key: str):
        return self._children[key]

    def has_child(self, key: str):
        return key in self._children

    __setitem__ = add_child
    __getitem__ = get_child
    __contains__ = has_child

    def __iter__(self):
        raise NotImplementedError


# ========================================================================= #
# Group                                                                     #
# ========================================================================= #


class Group(_ConfigObject):

    def __init__(
            self,
            named_nodes: Dict[str, Union['Group', 'Option']] = None,
    ):
        super().__init__()
        # add groups or options
        if named_nodes:
            for k, v in named_nodes.items():
                self.add_child(k, v)
            # we dont need to validate groups because we dont
            # directly support conversion to them yet.

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Override                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def add_child(self, key, node: Union['Group', 'Option']) -> Union['Group', 'Option']:
        if not isinstance(node, (Group, Option)):
            raise TypeError(f'Can only add {Group.__name__} and {Option.__name__} nodes to {Group.__name__}')
        # add!
        assert super().add_child(key, node) is node
        # we dont return the above because of type warnings
        return node

    # override to yield in order
    def __iter__(self):
        yield from self.groups
        yield from self.options

    def walk_descendant_options(self):
        for node in self.walk_descendants():
            if isinstance(node, Option):
                yield node

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Getters                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def root(self) -> 'Group':
        g = super().root
        assert isinstance(g, Group)
        return g

    @property
    def group(self) -> 'Group':
        return self

    @property
    def groups(self) -> Dict[str, 'Group']:
        return {k: v for k, v in self._children.items() if isinstance(v, Group)}

    @property
    def options(self) -> Dict[str, 'Option']:
        return {k: v for k, v in self._children.items() if isinstance(v, Option)}

    @property
    def group_keys(self) -> Tuple[str]:
        return self.keys

    @property
    def abs_group_path(self):
        return self.abs_path

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Package                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_package_keys(self, pkg_override=None) -> Tuple[str]:
        assert pkg_override is None, 'cannot override the groups package'
        return self.keys

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Groups & Options                                                      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_subgroup(self, key: str) -> 'Group':
        node = self.get_child(key)
        if not isinstance(node, Group):
            raise TypeError(f'node with key: {repr(key)} is not a {Group.__name__}')
        return node

    def del_subgroup(self, key: str) -> 'Group':
        assert self.has_subgroup(key)
        return self.del_child(key)

    def get_option(self, key: str) -> 'Option':
        node = self.get_child(key)
        if not isinstance(node, Option):
            raise TypeError(f'node with key: {repr(key)} is not a {Option.__name__}')
        return node

    def add_subgroup(self, key: str, value: 'Group') -> 'Group':
        if not isinstance(value, Group):
            raise TypeError(f'adding node with key: {repr(key)} is not a {Group.__name__}')
        return self.add_child(key, value)

    def add_option(self, key: str, value: 'Option') -> 'Option':
        if not isinstance(value, Option):
            raise TypeError(f'adding node with key: {repr(key)} is not a {Option.__name__}')
        return self.add_child(key, value)

    def del_option(self, key: str) -> 'Option':
        assert self.has_suboption(key)
        return self.del_child(key)

    def new_subgroup(self, key: str) -> 'Group':
        return self.add_subgroup(key, Group())

    def new_option(self, key: str, data=None, pkg=None, defaults=None) -> 'Option':
        return self.add_option(key, Option(
            data=data, pkg=pkg, defaults=defaults
        ))

    def has_suboption(self, key):
        if not self.has_child(key):
            return False
        if not isinstance(self.get_child(key), Option):
            return False
        return True

    def has_subgroup(self, key):
        if not self.has_child(key):
            return False
        if not isinstance(self.get_child(key), Group):
            return False
        return True

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Walk                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _get_keys_and_root(self, path: Union[str, List[str], Tuple[str]]):
        if isinstance(path, str):
            group_keys, is_relative = V.split_config_path(path)
            # get the group corresponding to the path - must handle relative & root paths
            root = (self if is_relative else self.root)
        else:
            group_keys, root = path, self
        return group_keys, root

    def get_child_recursive(self, path: Union[str, List[str], Tuple[str]], make_missing_as_groups=False) -> Union['Group', 'Option']:
        keys, current = self._get_keys_and_root(path)
        while keys:
            (key, *keys) = keys
            if isinstance(current, Option):
                raise KeyError(f'an option has no descendants, tired to visit child {repr(key)} of option {repr(current.abs_path)}')
            if not current.has_child(key):
                if not make_missing_as_groups:
                    raise KeyError(f'{current.__class__.__name__} {repr(current.abs_path)} does not have child {repr(key)}')
                else:
                    if not isinstance(current, Group):
                        raise KeyError(f'cannot make missing group {repr(key)} on option {repr(current.abs_path)} for path {repr(path)}')
                    current.new_subgroup(key)
            current = current.get_child(key)
        return current

    def get_group_recursive(self, path: Union[str, List[str], Tuple[str]], make_missing=False) -> 'Group':
        """
        This function checks if a path is relative or absolute.
        If the path is relative, it traverses from the current group.
        If the path is absolute, it traverses from the root group.

        NOTE: If the path is a *list* of keys,
              they are considered *relative*.

        Supports make_missing, to create any missing groups.
        """
        group = self.get_child_recursive(path, make_missing_as_groups=make_missing)
        if not isinstance(group, Group):
            raise KeyError(f'{group.__class__.__name__} config object at path {repr(path)} is not a group.')
        return group

    def get_option_recursive(self, path: Union[str, List[str], Tuple[str]]) -> 'Option':
        """
        This function is like get_group_recursive, checking for relative
        and absolute paths, except the last component of the path should
        be an option name.

        NOTE: If the path is a *list* of keys,
              they are considered *relative*.
        """
        (*group_keys, option_name), root = self._get_keys_and_root(path)
        group = self.get_group_recursive(group_keys, make_missing=False)
        if not group.has_suboption(option_name):
            raise KeyError(f'group {repr(group.abs_group_path)} does not have suboption {repr(option_name)}')
        return group.get_option(option_name)

    def has_group_recursive(self, path: Union[str, List[str], Tuple[str]]):
        try:
            self.get_group_recursive(path, make_missing=False)
            return True
        except KeyError:
            return False

    def has_option_recursive(self, path: Union[str, List[str], Tuple[str]]):
        try:
            self.get_option_recursive(path)
            return True
        except KeyError:
            return False

    def path_to_abs_keys(self, path: Union[str, List[str], Tuple[str]]) -> Tuple[str]:
        """
        convert a relative or abs path to abs keys
        """
        path, root = self._get_keys_and_root(path)
        return root.keys + tuple(path)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Strings                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self._children})'

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Merge                                                                 #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def absorb_children(self, group: 'Group', allow_replace_options=False) -> 'Group':
        # like dict.update, but modify right hand side...
        if not isinstance(group, Group):
            raise TypeError('group can only absorb other groups.')
        # -- this should be split out into a func
        if group.root is self.root:
            raise AssertionError('tried to absorb group from the same tree.')
        # recursive walk to check conflicts first, so that things arent modified if there is a failure
        # TODO: this is not very efficient, walks down tree each time to check...
        # -- this should maybe be split out into a func
        if not allow_replace_options:
            for r_option in group.walk_descendant_options():
                r_rel_keys = group.rel_keys_to(r_option)
                if self.has_option_recursive(r_rel_keys):
                    l_option = self.get_option_recursive(r_rel_keys)
                    raise KeyError(f'left group {repr(self.abs_path)} cannot absorb right group {repr(group.abs_path)} because conflicting left option was found at {repr(l_option.abs_path)} and right option at {repr(r_option.abs_path)}.')
        # do absorb!
        self._absorb_children(group, allow_replace_options=allow_replace_options)
        return self

    def _absorb_children(self, group: 'Group', allow_replace_options: bool):
        # absorb options
        for r_option_name, r_option in group.options.items():
            group.del_option(r_option_name)
            # remove left option if allowed
            if self.has_suboption(r_option_name):
                if allow_replace_options:
                    self.del_option(r_option_name)
                else:
                    raise AssertionError('This is a bug and should never happen! Check that no conflicts exist first!')
            # add right option to left
            self.add_option(r_option_name, r_option)
        # absorb subgroups
        for r_group_name, r_group in group.groups.items():
            group.del_subgroup(r_group_name)
            if self.has_subgroup(r_group_name):
                # recursively absorb if same group already exists in left
                self.get_subgroup(r_group_name)._absorb_children(r_group, allow_replace_options=allow_replace_options)
            else:
                # add right group to left
                self.add_subgroup(r_group_name, r_group)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Debug                                                                 #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def debug_tree_print(
            self, colors=True, show_options=True, full_option_path=True,
            full_group_path=True, show_defaults=False
    ):
        from eunomia.config._config_debug import debug_tree_print
        debug_tree_print(
            self, colors=colors, show_options=show_options, full_option_path=full_option_path,
            full_group_path=full_group_path, show_defaults=show_defaults
        )


# ========================================================================= #
# Option                                                                    #
# ========================================================================= #


class Option(_ConfigObject):

    def __init__(
            self,
            data: dict = None,
            pkg: str = None,
            defaults: list = None,
    ):
        super().__init__()
        # extract components
        self._data = V.validate_option_data(data)
        self._pkg = V.validate_option_package(pkg)
        self._defaults = V.validate_option_defaults(defaults, allow_config_nodes=True)

    @property
    def pkg(self) -> str:
        return self._pkg

    @property
    def data(self) -> dict:
        return self._data.copy()

    @property
    def defaults(self) -> list:
        return self._defaults

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # getters                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def root(self) -> 'Group':
        assert self.has_parent
        g = super().root
        assert isinstance(g, Group)
        return g

    @property
    def group(self) -> Group:
        assert self.has_parent
        g = self.parent
        assert isinstance(g, Group)
        return g

    @property
    def group_keys(self) -> Tuple[str]:
        return self.group.keys

    @property
    def abs_group_path(self) -> str:
        return self.group.abs_group_path

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Package                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_package_keys(self, pkg_override: str = None) -> Tuple[str]:
        pkg = self._pkg if (pkg_override is None) else pkg_override
        # resolve special keys
        if pkg in K.ALL_PACKAGE_ALIASES:
            if pkg == K.PKG_ROOT: return tuple()
            elif pkg == K.PKG_GROUP: return self.group_keys
            elif pkg == K.PKG_OPTION: return self.keys
            else: raise AssertionError('it is a bug and should never happen.')
        # split
        keys, is_relative = V.split_package_path(pkg)
        # resolve relative -- NOTE! RELATIVE from the GROUP not the option
        if is_relative:
            keys = self.group_keys + keys
        return keys

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # getters - data                                                        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    class _ReplaceStrings(RecursiveTransformer):
        def _transform_str(self, value):
            return SubNode(value)
        def __transform_default__(self, value):
            return value
        def _transform_dict_key(self, key):
            # do not allow interpolation on keys
            if isinstance(key, str):
                return key
            elif isinstance(key, ConfigNode):
                raise TypeError(f'keys in configs cannot be config nodes: {key}')
            return super()._transform_dict_key(key)

    def get_unresolved_defaults(self) -> List['Default']:
        from eunomia.config._default import Default
        return [Default(d) for d in self._defaults]

    def get_unresolved_package(self):
        return self._ReplaceStrings().transform(self._pkg)

    def get_unresolved_data(self):
        return self._ReplaceStrings().transform(self._data)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Children - disabled for the option node                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    add_child = None
    get_child = None
    __setitem__ = None
    __getitem__ = None
    __iter__ = ().__iter__

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Strings                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}(data={repr(self._data)}, pkg={repr(self._pkg)}, defaults={repr(self._defaults)})'


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import re
from typing import Union as _Union
from eunomia.config._config import _ConfigObject, Group, Option
from eunomia.config.nodes import ConfigNode
from eunomia.config import validate as V


# ========================================================================= #
# debug tree                                                                #
# ========================================================================= #


def debug_tree_print(root: _Union[Group, Option], colors=True, show_options=True, full_option_path=True, full_group_path=True, show_defaults=True):
    print(debug_tree_str(
        root=root, colors=colors, show_options=show_options, full_option_path=full_option_path,
        full_group_path=full_group_path, show_defaults=show_defaults
    ))


def debug_tree_str(root: _Union[Group, Option], colors=True, show_options=True, full_option_path=True, full_group_path=True, show_defaults=True):
    from attr import dataclass

    @dataclass
    class _WalkObj:
        node: _ConfigObject
        visited: bool
        is_last: bool
        @property
        def is_group(self):
            return isinstance(self.node, Group)
        @property
        def has_groups_after(self):
            return self.node.has_parent and isinstance(self.node.parent, Group) and bool(self.node.parent.groups)
        @property
        def key(self):
            return (self.is_group, self.visited, self.is_last, self.has_groups_after)
        def __str__(self):
            return f":{''.join(map(str, map(int, self.key)))}"

    def _is_last_iter(items):
        yield from ((item, i == len(items ) -1) for i, item in enumerate(items))

    def _walk():
        def _recurse(node, is_last, stack):
            stk = stack + [_WalkObj(node, False, is_last)]
            yield stk
            stk[-1].visited = True
            if isinstance(node, Group):
                if show_options:
                    for o, l in _is_last_iter(node.options.values()):
                        yield from _recurse(o, l, stk)
                for g, l in _is_last_iter(node.groups.values()):
                    yield from _recurse(g, l, stk)
        return _recurse(root, True, [])

    if colors:
        nG, nO, S, G, O, R = '\033[35m', '\033[33m', '\033[90m', '\033[95m', '\033[93m', '\033[0m'
    else:
        nG, nO, S, G, O, R = '', '', '', '', '', ''

    TREE = {
        (1, 1, 1, 0): f'   ',           # group,  visited,   last,  has groups after
        (1, 1, 1, 1): f'   ',           # group,  visited,   last,  no groups after
        (1, 0, 1, 0): f'   ',           # group,  unvisited, last,  has groups after
        (1, 1, 0, 1): f' {S}│{R} ',     # group,  visited,   inner, no groups after
        (1, 0, 0, 1): f' {S}├{G}─{R}',  # group,  unvisited, inner, no groups after
        (1, 0, 1, 1): f' {S}╰{G}─{R}',  # group,  unvisited, last,  no groups after
        (0, 0, 0, 1): f' {S}│{R} ',     # option, unvisited, last,  has groups after
        (0, 0, 1, 1): f' {S}├{O}╌{R}',  # option, unvisited, last,  no groups after
        (0, 0, 0, 0): f' {S}├{O}╌{R}',  # option, unvisited, inner, has groups after
        (0, 0, 1, 0): f' {S}╰{O}╌{R}',  # option, unvisited, last,  has groups after
    }

    # get correct line length without ansi colors
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')

    lengths, lines, items = {}, [], []
    for stack in _walk():
        (*_, item) = stack
        tree = ''.join(TREE.get(o.key, f'ERR{o}') for o in stack[1:])
        if item.is_group:
            # GROUP
            keys = [f'/{k}' for k in (item.node.keys if item.node.keys else ('',))]
            name = f"{nG}{keys[-1]}{R}"
            if full_group_path:
                name = f"{S}{''.join(keys[:-1])}{R}" + name
        else:
            # OPTION
            name = f"{nO}{item.node.key}{R}"
            if full_option_path:
                name = f"{S}{V.keys_as_abs_config_path(item.node.keys[:-1])}:{R} " + name
        # append for next iterations
        line = f'{tree} {name}'
        # get max line length
        lengths.setdefault(item.node.parent, 0)
        lengths[item.node] = len(ansi_escape.sub('', line))
        lengths[item.node.parent] = max(lengths[item.node.parent], lengths[item.node])
        # append
        lines.append(line)
        items.append(item)

    def get_defaults(option):
        for default in option.get_unresolved_defaults():
            default = default._default
            if isinstance(default, Option):
                yield f'{default.abs_group_path}: {default.key}'
            elif isinstance(default, dict):
                ((k, v),) = default.items()
                if isinstance(k, (Group, Option)): k = k.abs_path
                if isinstance(v, (Group, Option)): v = v.abs_path
                yield f'{k}: {v}'
            elif isinstance(default, ConfigNode):
                yield str(default)
            else:
                yield str(default)

    if show_defaults:
        new_lines = []
        for line, item in zip(lines, items):
            if not item.is_group:
                defaults = list(get_defaults(item.node))
                if defaults:
                    padding = lengths[item.node.parent] - lengths[item.node]
                    line = f'{line}{" "*padding} {S}[{", ".join(defaults)}]{R}'
            new_lines.append(line)
        lines = new_lines

    return '\n'.join(lines)


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



from typing import Union

from eunomia.config import Group as Group, Option


# ========================================================================= #
# Config Backend                                                            #
# ========================================================================= #


class Backend(object):

    """
    A backend is an instance of various settings needed to load
    a group or option from some value.
    """

    @classmethod
    @property
    def GROUP_TYPE(self):
        raise NotImplementedError

    @classmethod
    @property
    def OPTION_TYPE(self):
        raise NotImplementedError

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Load                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def load_group(self, value: GROUP_TYPE):
        if not isinstance(value, self.GROUP_TYPE):
            raise TypeError(f'group value must be of type: {self.GROUP_TYPE}')
        group = self._load_group(value)
        assert isinstance(group, Group), f'loaded group must be a: {Group}'
        return group

    def load_option(self, value: OPTION_TYPE):
        if not isinstance(value, self.OPTION_TYPE):
            raise TypeError(f'option value must be of type: {self.OPTION_TYPE}')
        option = self._load_option(value)
        assert isinstance(option, Option), f'loaded option must be a: {Option}'
        return option

    def _load_group(self, value: GROUP_TYPE) -> Group:
        raise NotImplementedError

    def _load_option(self, value: OPTION_TYPE) -> Option:
        raise NotImplementedError

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Dump                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def dump(self, config: Union[Group, Option]):
        if isinstance(config, Group):
            return self.dump_group(config)
        elif isinstance(config, Option):
            return self.dump_option(config)
        else:
            raise TypeError(f'unsupported config type: {type(config)}')

    def dump_group(self, group: Group):
        if not isinstance(group, Group):
            raise TypeError(f'group must be a: {Group}')
        value = self._dump_group(group)
        assert isinstance(value, self.GROUP_TYPE), f'dumped group value must be of type: {self.GROUP_TYPE}'
        return value

    def dump_option(self, option: Option):
        if not isinstance(option, Option):
            raise TypeError(f'option must be a: {Option}')
        value = self._dump_option(option)
        assert isinstance(value, self.OPTION_TYPE), f'dumped option value must be of type: {self.OPTION_TYPE}'
        return value

    def _dump_group(self, group: Group):
        raise NotImplementedError

    def _dump_option(self, option: Option):
        raise NotImplementedError


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import keyword as _keyword
from typing import Union as _Union, Tuple as _Tuple, Dict as _Dict, List as _List

from eunomia.config import keys as _K


# ========================================================================= #
# Keys                                                                      #
# ========================================================================= #


def validate_identifier(key) -> str:
    if not isinstance(key, str):
        raise TypeError(f'identifier is not a string type: {type(key)}')
    if not key:
        raise ValueError(f'identifier cannot be empty: {repr(key)}')
    if not str.isidentifier(key):
        raise ValueError(f'identifier is not a valid python identifier: {repr(key)}')
    if _keyword.iskeyword(key):
        raise ValueError(f'identifier is a python keyword: {repr(key)}')
    return key


def validate_identifier_list(keys) -> list:
    if not isinstance(keys, list):
        raise TypeError(f'identifier list must be a list type: {type(keys)}')
    for k in keys:
        validate_identifier(k)
    return keys


def validate_config_identifier(key) -> str:
    key = validate_identifier(key)
    if key in _K.RESERVED_KEYS:
        raise ValueError(f'identifier is eunomia reserved key: {repr(key)}')
    if key.startswith('__') and key.endswith('__'):
        raise ValueError(f'identifier is reserved: {repr(key)}')
    return key


def validate_config_identifier_list(keys) -> list:
    if not isinstance(keys, list):
        raise TypeError(f'identifier list must be a list type: {type(keys)}')
    for k in keys:
        validate_config_identifier(k)
    return keys


# ========================================================================= #
# Values                                                                    #
# ========================================================================= #


def _split_path(path: str, sep: str) -> (str, bool):
    # check and remove prefix
    is_prefixed = False
    if path.startswith(sep):
        is_prefixed, path = True, path[len(sep):]
    # split path
    if path:
        split = str.split(path, sep)
    else:
        split = []
    # check identifiers
    return validate_config_identifier_list(split), is_prefixed


def split_package_path(path: str) -> (str, bool):
    if path in _K.ALL_PACKAGE_ALIASES:
        raise RuntimeError('special package keys should be handled separately')
    keys, is_relative = _split_path(path, '.')
    return keys, is_relative


def split_config_path(path: str) -> (str, bool):
    keys, is_not_relative = _split_path(path, '/')
    return keys, not is_not_relative


# ========================================================================= #
# Paths                                                                     #
# ========================================================================= #


def validate_package_path(path) -> str:
    if not isinstance(path, str):
        raise TypeError(f'package path is not of type string: {type(path)}')
    # check for aliases
    if path in _K.ALL_PACKAGE_ALIASES:
        return path
    # try split
    split_package_path(path)
    return path


def validate_config_path(path) -> str:
    if not isinstance(path, str):
        raise TypeError(f'config path is not of type string: {path}')
    # try split
    split_config_path(path)
    return path


# ========================================================================= #
# config values                                                             #
# ========================================================================= #


def validate_option_data(data) -> dict:
    if data is None:
        return {}

    from eunomia.config.nodes import ConfigNode

    if isinstance(data, ConfigNode):
        raise TypeError(f'option data object can never be a config node: {repr(data)}')
    if not isinstance(data, dict):
        raise TypeError(f'option data object must be of type dict: {repr(data)}')
    return _validate_option_data(data)


def _validate_option_data(value):
    if value is None:
        pass
    elif isinstance(value, dict):
        for k, v in value.items():
            validate_config_identifier(k)
            _validate_option_data(v)
    elif isinstance(value, (list, tuple, set)):
        for v in value:
            _validate_option_data(v)
    elif isinstance(value, (int, float, str)):
        pass
    else:
        raise TypeError(f'unsupported data value type: {type(value)} with value: {repr(value)}')
    return value


def validate_option_package(pkg) -> str:
    if pkg is None:
        return _K.DEFAULT_PKG

    from eunomia.config.nodes import ConfigNode

    if isinstance(pkg, ConfigNode):
        raise TypeError(f'option package can never be a config node: {repr(pkg)}')
    try:
        return validate_package_path(pkg)
    except Exception as e:
        raise ValueError(f'option package is invalid: {repr(pkg)}').with_traceback(e.__traceback__)


def validate_option_defaults(defaults, allow_config_nodes=False) -> list:
    if defaults is None:
        return []

    from eunomia.config.nodes import ConfigNode

    if isinstance(defaults, ConfigNode):
        raise TypeError(f'option defaults can never directly be a config node, only its items: {repr(defaults)}')
    if not isinstance(defaults, list):
        raise TypeError('defualts must be a list of values')

    try:
        for default in defaults:
            validate_defaults_item(default, allow_config_nodes=allow_config_nodes)
    except Exception as e:
        raise ValueError(f'option defaults is invalid: {repr(defaults)}').with_traceback(e.__traceback__)

    return defaults


def validate_option_type(typ) -> str:
    if typ is None:
        return _K.TYPE_OPTION

    from eunomia.config.nodes import ConfigNode

    if isinstance(typ, ConfigNode):
        raise TypeError(f'option type can never be a config node: {repr(typ)}')
    if not isinstance(typ, str):
        raise TypeError(f'option type must be of type string: {repr(typ)}')
    if typ != _K.TYPE_OPTION:
        raise ValueError(f'option  type must be: {repr(typ)}')
    return typ


# ========================================================================= #
# Options                                                                   #
# ========================================================================= #


def validate_defaults_item(default, allow_config_nodes=False) -> _Union[str, _Dict[str, str], _Dict[str, _List[str]]]:
    from eunomia.config import Group, Option
    from eunomia.config.nodes import ConfigNode

    # we cant resolve options or defaults here in case the tree is modified after adding!
    # we can only resolve them at merge time!
    if isinstance(default, str):
        if default != _K.OPT_SELF:
            validate_config_path(default)
    elif isinstance(default, (Group, Option, ConfigNode)):
        if not allow_config_nodes:
            raise TypeError(f'{default.__class__.__name__} have been disabled as the group key for default entry. Invalid: {default}')
    elif isinstance(default, dict):
        # split
        if len(default) != 1:
            raise ValueError(f'if defaults entry is a dictionary it must have one and only one key. Invalid: {default}')
        ((group, opts),) = default.items()

        # check key
        if isinstance(group, str):
            validate_config_path(group)
        elif isinstance(group, Group):
            if not allow_config_nodes:
                raise TypeError(f'{group.__class__.__name__} have been disabled as the group key for default entry. Invalid: {group}')
        elif isinstance(group, ConfigNode):
            raise TypeError(f'substitution is not supported for the group key of a defaults entry: {group}')
        else:
            raise TypeError(f'unsupported defaults entry group key type. Invalid: {group}')

        # check value
        if isinstance(opts, str):
            if opts != _K.OPT_GLOB:
                validate_config_path(opts)
        elif isinstance(opts, (Group, Option, ConfigNode)):
            if not allow_config_nodes:
                raise TypeError(f'{opts.__class__.__name__} have been disabled as the option values for default entry. Invalid: {opts}')
        elif isinstance(opts, list):
            for opt in opts:
                if isinstance(opt, str):
                    validate_config_path(opt)
                elif isinstance(opt, Option):
                    if not allow_config_nodes:
                        raise TypeError(f'{opts.__class__.__name__} have been disabled as the item in option value list for default entry. Invalid: {opts}')
                else:
                    raise TypeError(f'unsupported item in defaults entry value list type. Invalid: {opts}')
        else:
            raise TypeError(f'unsupported defaults entry group value type {type(opts)} with value {opts}')
        # done!
    else:
        raise TypeError(f'default entry is invalid type {type(default)}, can only be string paths, {Group.__name__}s or {Option.__name__}s, or a dictionary containing those.')


def keys_as_abs_config_path(keys: _Union[_Tuple[str], _List[str]]) -> str:
    return validate_config_path('/' + '/'.join(keys))


def keys_as_abs_pkg_path(keys: _Union[_Tuple[str], _List[str]]) -> str:
    return validate_package_path('.'.join(keys))


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
